import  numpy as np
from scipy.integrate import odeint

__all__ = [
    'solve_ode_one_site_insolution',
    'signal_ode_one_site_insolution',
    'ode_induced_fit_insolution',
    'solve_ode_induced_fit_insolution',
    'signal_ode_induced_fit_insolution',
    'ode_conformational_selection_insolution',
    'solve_ode_conformational_selection_insolution',
    'signal_ode_conformational_selection_insolution',
    'get_initial_concentration_conformational_selection',
    'get_kobs_induced_fit',
    'get_kobs_conformational_selection'
]

def solve_ode_one_site_insolution(t, koff, Kd, a_total, b_total, t0=0):
    """
    Analytical solution for the one binding site model in solution.

    Assumes that the initial concentration of the complex is zero.

    The ODE  dAB/dt = kon·(A_tot − AB)·(B_tot − AB) − koff·AB  is a Riccati
    equation whose two fixed points x₁ (physical) and x₂ are the roots of

        kon·x² − (kon·(A+B) + koff)·x + kon·A·B = 0

    Starting from AB(0) = 0 the closed-form solution is

        AB(t) = x₁·(1 − e^{λt}) / (1 − (x₁/x₂)·e^{λt})

    with λ = −kon·√(p² − 4·A·B), p = A + B + Kd.

    Parameters
    ----------
    t : np.ndarray
        Time points to calculate the complex concentration.
    koff : float
        Off rate constant.
    Kd : float
        Equilibrium dissociation constant.
    a_total : float
        Total concentration of the ligand.
    b_total : float
        Total concentration of the receptor.
    t0 : float, optional
        Initial time, default is 0.

    Returns
    -------
    np.ndarray
        Complex concentration over time.
    """
    t = t + t0

    kon = koff / Kd

    p = a_total + b_total + Kd
    disc = p * p - 4 * a_total * b_total   # always ≥ Kd² > 0
    D = np.sqrt(disc)

    x1 = (p - D) * 0.5          # physical steady state  (≤ min(A,B))
    x2 = (p + D) * 0.5          # unphysical root

    lam = -kon * D               # negative eigenvalue → decay

    exp_lt = np.exp(lam * t)
    AB = x1 * (1.0 - exp_lt) / (1.0 - (x1 / x2) * exp_lt)

    return AB

def signal_ode_one_site_insolution(t, koff, Kd, a_total, b_total, t0=0, signal_a=0, signal_b=0, signal_complex=0):
    """
    Solve the ODE for the one binding site model and compute the signal.

    Parameters
    ----------
    t : np.ndarray
        Time.
    koff : float
        Off rate constant.
    Kd : float
        Equilibrium dissociation constant.
    a_total : float
        Total concentration of the ligand.
    b_total : float
        Total concentration of the receptor.
    t0 : float, optional
        Initial time, default is 0.
    signal_a : float, optional
        Signal for the ligand, default is 0.
    signal_b : float, optional
        Signal for the receptor, default is 0.
    signal_complex : float, optional
        Signal for the complex, default is 0.

    Returns
    -------
    np.ndarray
        Signal over time.
    """
    complex_conc = solve_ode_one_site_insolution(t, koff, Kd, a_total, b_total, t0)

    signal = signal_a * (a_total - complex_conc) + signal_b * (b_total - complex_conc) + signal_complex * complex_conc

    return signal


def ode_induced_fit_insolution(t, y, k1, k_minus1, k2, k_minus2, E_tot, S_tot):
    """
    Reduced ODEs for induced fit model using conservation of E and S.
    
    Parameters
    ----------
    t : float
        Time.
    y : list
        Concentrations of E·S and ES.
    k1 : float
        Rate constant for E + S -> E·S.
    k_minus1 : float
        Rate constant for E·S -> E + S.
    k2 : float
        Rate constant for E·S -> ES.
    k_minus2 : float
        Rate constant for ES -> E·S.
    E_tot : float
        Total concentration of the enzyme.
    S_tot : float
        Total concentration of the substrate.
        
    Returns
    -------
    list
        [dE_S, dES] - Rate of change for concentrations of E·S and ES.
    """
    E_S, ES = y
    E = E_tot - E_S - ES
    S = S_tot - E_S - ES

    dE_S = k1 * E * S - k_minus1 * E_S - k2 * E_S + k_minus2 * ES
    dES  = k2 * E_S   - k_minus2 * ES

    return [dE_S, dES]

def solve_ode_induced_fit_insolution(t, y0, k1, k_minus1, k2, k_minus2, E_tot, S_tot, t0=0):
    """
    Solve the reduced ODE for the induced fit model in solution.
    
    Parameters
    ----------
    t : np.ndarray
        Time points to calculate the complex concentration.
    y0 : list
        Initial concentrations of E·S and ES.
    k1 : float
        Rate constant for E + S -> E·S.
    k_minus1 : float
        Rate constant for E·S -> E + S.
    k2 : float
        Rate constant for E·S -> ES.
    k_minus2 : float
        Rate constant for ES -> E·S.
    E_tot : float
        Total concentration of the enzyme.
    S_tot : float
        Total concentration of the substrate.
    t0 : float, optional
        Initial time, default is 0.
        
    Returns
    -------
    np.ndarray
        Solution of the ODE, contains the concentration of E, S, E·S and ES.
    """
    t = t + t0
    sol = odeint(ode_induced_fit_insolution, y0, t,
                 args=(k1, k_minus1, k2, k_minus2, E_tot, S_tot), tfirst=True)
    E_S = sol[:, 0]
    ES  = sol[:, 1]
    # Include the concentrations of E and S in the output
    E = E_tot - E_S - ES
    S = S_tot - E_S - ES
    out = np.vstack((E, S, E_S, ES))  # Stack E, S, E_S, ES

    return out

def signal_ode_induced_fit_insolution(t, y, k1, k_minus1, k2, k_minus2, E_tot, S_tot,
                                      t0=0, signal_E=0, signal_S=0, signal_ES_int=0, signal_ES=0):
    """
    Solve the reduced ODE for the induced fit model and compute the signal.
    
    Parameters
    ----------
    t : np.ndarray
        Time.
    y : list
        Concentrations of E·S and ES.
    k1 : float
        Rate constant for E + S -> E·S.
    k_minus1 : float
        Rate constant for E·S -> E + S.
    k2 : float
        Rate constant for E·S -> ES.
    k_minus2 : float
        Rate constant for ES -> E·S.
    E_tot : float
        Total concentration of the enzyme.
    S_tot : float
        Total concentration of the substrate.
    t0 : float, optional
        Initial time, default is 0.
    signal_E : float, optional
        Signal for E, default is 0.
    signal_S : float, optional
        Signal for S, default is 0.
    signal_ES_int : float, optional
        Signal for E·S, default is 0.
    signal_ES : float, optional
        Signal for ES, default is 0.
        
    Returns
    -------
    np.ndarray
        Signal over time.
    """
    species = solve_ode_induced_fit_insolution(t, y, k1, k_minus1, k2, k_minus2, E_tot, S_tot, t0)
    signal = signal_E * species[0] + signal_S * species[1] + signal_ES_int * species[2] + signal_ES * species[3]
    return signal


def ode_conformational_selection_insolution(t, y, k1, k_minus1, k2, k_minus2, E_tot, S_tot):

    """
    Reduced ODEs for conformational selection using conservation:
      E_tot = E1 + E2 + E2S
      S_tot = S + E2S

    State vector y: [E2, E2S]

    Returns: [dE2_dt, dE2S_dt]
    """

    E2, E2S = y
    E1 = E_tot - E2 - E2S
    S  = S_tot - E2S

    dE2  = k1 * E1 - k_minus1 * E2 - k2 * E2 * S + k_minus2 * E2S
    dE2S = k2 * E2 * S - k_minus2 * E2S

    return [dE2, dE2S]

def solve_ode_conformational_selection_insolution(t, y, k1, k_minus1, k2, k_minus2, E_tot,S_tot,t0=0):
    """
    Solve ODE for the conformational selection model.

    Parameters
    ----------
    t : np.ndarray
        Time points to calculate the complex concentration.
    y : list
        Initial concentrations of E2 and E2S.
    k1 : float
        Rate constant for E1 -> E2.
    k_minus1 : float
        Rate constant for E2 -> E1.
    k2 : float
        Rate constant for E2 + S -> E2S.
    k_minus2 : float
        Rate constant for E2S -> E2 + S.
    E_tot : float
        total protein concentration
    S_tot : float
        total substrate concentration
    t0 : float, optional
        Initial time, default is 0.

    Returns
    -------
    np.ndarray
        Solution of the ODE, contains the concentration of E1, E2, S, E2S.
    """
    t = t + t0

    sol = odeint(ode_conformational_selection_insolution, y, t,
                 args=(k1, k_minus1, k2, k_minus2, E_tot, S_tot), tfirst=True)

    E2  = sol[:, 0]
    E2S = sol[:, 1]
    E1  = E_tot - E2 - E2S
    S   = S_tot - E2S

    return np.vstack((E1, E2, S, E2S))

def signal_ode_conformational_selection_insolution(
        t, y,
        k1, k_minus1,
        k2, k_minus2,
        E_tot, S_tot,
        t0=0, signal_E1=0, signal_E2=0, signal_S=0, signal_E2S=0):
    """
    Solve the ODE for the conformational selection model and compute the signal.
    
    Parameters
    ----------
    t : np.ndarray
        Time.
    y : list
        Concentrations of E2 and E2S.
    k1 : float
        Rate constant for E1 -> E2 (aka kc).
    k_minus1 : float
        Rate constant for E2 -> E1 (aka krev).
    k2 : float
        Rate constant for E2 + S -> E2S (aka kon).
    k_minus2 : float
        Rate constant for E2S -> E2 + S (aka koff).
    E_tot : float
        total protein concentration
    S_tot : float
        total substrate concentration
    t0 : float, optional
        Initial time, default is 0.
    signal_E1 : float, optional
        Signal for E1, default is 0.
    signal_E2 : float, optional
        Signal for E2, default is 0.
    signal_S : float, optional
        Signal for S, default is 0.
    signal_E2S : float, optional
        Signal for E2S, default is 0.
        
    Returns
    -------
    np.ndarray
        Signal over time.
    """
    species = solve_ode_conformational_selection_insolution(t, y, k1, k_minus1, k2, k_minus2, E_tot, S_tot, t0)

    signal = signal_E1 * species[0] + signal_E2 * species[1] + signal_S * species[2] + signal_E2S * species[3]

    return signal

def get_initial_concentration_conformational_selection(total_protein, k1, k_minus1):
    """
    Obtain the equilibrium concentrations of E1 and E2 for the conformational selection model.
    
    Assumes absence of ligand.
    
    Parameters
    ----------
    total_protein : float
        Total concentration of the protein.
    k1 : float
        Rate constant for E1 -> E2.
    k_minus1 : float
        Rate constant for E2 -> E1.
        
    Returns
    -------
    list
        [E1, E2] - Equilibrium concentrations of E1 and E2.
    """
    E1 = total_protein / (1 + k1 / k_minus1)
    E2 = total_protein - E1

    return [E1, E2]

def get_kobs_induced_fit(tot_lig, tot_prot, kr, ke, kon, koff, dominant=True):
    """
    Calculate the observed rate constant (relaxation rate) for the induced fit model in solution.
    
    Parameters
    ----------
    tot_lig : float
        Total concentration of the ligand.
    tot_prot : float
        Total concentration of the protein.
    kr : float
        Rate constant for E2S -> E1S.
    ke : float
        Rate constant for E1S -> E2S.
    kon : float
        Rate constant for E1 + S -> E1S.
    koff : float
        Rate constant for E1S -> E1 + S.
    dominant : bool, optional
        If True, calculate the dominant relaxation rate, otherwise calculate the non-dominant relaxation rate, default is True.
        
    Returns
    -------
    float
        Observed rate constant (in 1/s).
    """
    kd_app = koff * ke / (kon * (ke + kr))
    delta  = np.sqrt((tot_lig - tot_prot + kd_app) ** 2 + 4 * tot_prot * kd_app)
    gamma  = -ke - kr + koff + kon * (delta - kd_app)

    kobs = ke + kr + 0.5 * gamma + (-1*dominant * 0.5 * np.sqrt(gamma ** 2 + 4 * koff * kr))

    return kobs  # units are 1/s

def get_kobs_conformational_selection(tot_lig, tot_prot, kr, ke, kon, koff, dominant=True):
    """
    Calculate the observed rate constant (relaxation rate) for the conformational selection model in solution.
    
    Parameters
    ----------
    tot_lig : float
        Total concentration of the ligand.
    tot_prot : float
        Total concentration of the protein.
    kr : float
        Rate constant for E2 -> E1.
    ke : float
        Rate constant for E1 -> E2.
    kon : float
        Rate constant for E2 + S -> E2S.
    koff : float
        Rate constant for E2S -> E2 + S.
    dominant : bool, optional
        If True, calculate the dominant relaxation rate, otherwise calculate the non-dominant relaxation rate, default is True.
        
    Returns
    -------
    float
        Observed rate constant (in 1/s).
    """
    kd_app = koff * (ke + kr) / (kon * ke)
    delta = np.sqrt((tot_lig - tot_prot + kd_app)**2 + 4 * tot_prot * kd_app)
    beta = 2 * kr * (2 * ke - koff - koff * (delta - tot_lig + tot_prot) / kd_app)
    alpha = kr - ke + koff * ((2 * ke + kr) * delta + kr * (tot_lig - tot_prot - kd_app)) / (2 * ke * kd_app)


    kobs = ke + 0.5 * alpha + (-1*dominant* 0.5 * np.sqrt(alpha**2 + beta))

    return kobs  # units are 1/s


import itertools
import numpy as np
import pandas as pd

from .utils.fitting_surface import (
    fit_steady_state_one_site,
    steady_state_one_site_asymmetric_ci95,
    fit_one_site_association,
    fit_one_site_dissociation,
    fit_one_site_assoc_and_disso,
    fit_induced_fit_sites_assoc_and_disso,
    fit_one_site_assoc_and_disso_ktr,
    one_site_assoc_and_disso_asymmetric_ci95,
    one_site_assoc_and_disso_asymmetric_ci95_koff,
    get_smax_upper_bound_factor
)

from .utils.fitting_general import (
    fit_single_exponential,
    re_fit,
    re_fit_2
)

from .utils.processing      import (
    subset_data,
    concat_signal_lst,
    expand_parameter_list
)

from .utils.math import  get_desired_rss

from .fitter import KineticsFitterGeneral

class KineticsFitter(KineticsFitterGeneral):
    """
    A class used to fit kinetics data with shared thermodynamic parameters.

    Parameters
    ----------
    time_assoc_lst : list
        List of time points for the association signals, one per replicate.
    association_signal_lst : list
        List of association signals, one per replicate.
    lig_conc_lst : list
        List of ligand concentrations, one per replicate.
    time_diss_lst : list, optional
        List of time points for the dissociation signals, one per replicate.
    dissociation_signal_lst : list, optional
        List of dissociation signals, one per replicate.
    smax_id : list, optional
        List containing the Smax IDs (maximum amplitude identifiers).
    name_lst : list, optional
        List of experiment names.
    is_single_cycle : bool, optional
        Whether the experiment is a single cycle kinetics experiment.

    Attributes
    ----------
    names : list
        List of experiment names.
    assoc_lst : list
        List of association signals, one per replicate.
        Each signal is a numpy matrix of size n*m where n is the number of time points and m
        is the number of ligand concentrations.
    disso_lst : list
        List of dissociation signals, one per replicate.
        Each signal is a numpy matrix of size n*m where n is the number of time points and m
        is the number of ligand concentrations (different from zero).
    lig_conc_lst : list
        List of ligand concentrations, one per replicate.
    time_assoc_lst : list
        List of time points for the association signals, one per replicate.
    time_disso_lst : list
        List of time points for the dissociation signals, one per replicate.
    is_single_cycle : bool
        Whether the experiment is a single cycle kinetics experiment.
    signal_ss : list
        List of steady state signals, one per replicate.
    signal_ss_fit : list
        List of steady state fitted signals, one per replicate.
    signal_assoc_fit : list
        List of association kinetics fitted signals, one per replicate.
    signal_disso_fit : list
        List of dissociation kinetics fitted signals, one per replicate.
    lig_conc_lst_per_id : list
        Ligand concentrations per Smax ID.
    smax_guesses_unq : list
        Smax guesses per association signal.
    smax_guesses_shared : list
        Smax guesses per Smax ID.
    fit_params_kinetics : pd.DataFrame
        DataFrame with the fitted parameters for the association/dissociation kinetics.
    fit_params_ss : pd.DataFrame
        DataFrame with the values of the fitted parameters - steady state.
    """

    def __init__(self,time_assoc_lst,association_signal_lst,lig_conc_lst,
                 time_diss_lst=None,dissociation_signal_lst=None,
                 smax_id=None,name_lst=None,is_single_cycle=False):
        """
        Initialize the KineticsFitter class.
        
        Parameters
        ----------
        time_assoc_lst : list
            List of time points for the association signals, one per replicate.
        association_signal_lst : list
            List of association signals, one per replicate.
        lig_conc_lst : list
            List of ligand concentrations, one per replicate.
        time_diss_lst : list, optional
            List of time points for the dissociation signals, one per replicate.
        dissociation_signal_lst : list, optional
            List of dissociation signals, one per replicate.
        smax_id : list, optional
            List containing the Smax IDs (maximum amplitude identifiers).
        name_lst : list, optional
            List of experiment names.
        is_single_cycle : bool, optional
            Whether the experiment is a single cycle kinetics experiment, default is False.
        """

        super().__init__

        self.names            = name_lst
        self.assoc_lst        = association_signal_lst
        self.disso_lst        = dissociation_signal_lst
        self.lig_conc_lst     = lig_conc_lst
        self.time_assoc_lst   = time_assoc_lst
        self.time_disso_lst   = time_diss_lst
        self.is_single_cycle  = is_single_cycle

        self.lig_conc_lst_per_id = None  # Ligand concentrations per Smax ID
        self.smax_guesses_unq    = None  # Smax guesses per association signal
        self.smax_guesses_shared = None  # Smax guesses per Smax ID

        self.Kd_ss = None

        # If smax_id is None, set it to a list of zeros
        if smax_id is None:
            smax_id = [0 for _ in range(len(association_signal_lst))]

        # We need to rearrange smax_id to start at 0, then 1, then 2, etc.
        smax_id_unq, idx = np.unique(smax_id, return_index=True)
        smax_id_unq = smax_id_unq[np.argsort(idx)]

        smax_id_new = []
        for i,unq in enumerate(smax_id_unq):
            smax_id_new += [i for _ in range(len(np.where(smax_id == unq)[0]))]

        self.smax_id = smax_id_new

        self.signal_assoc_fit = None
        self.signal_disso_fit = None
        self.signal_ss_fit    = None

        self.Smax_upper_bound_factor = 1e2  # Normal values for lower than micromolar affinity

    def get_steady_state(self):

        """
        Calculate the steady state signal and group it by Smax ID.
        
        This function calculates the steady state signal for each association curve
        and groups the signals by Smax ID. It also calculates the Smax guesses
        for each association signal.
        
        Returns
        -------
        None
            Updates the following instance attributes:
            - signal_ss: List of steady state signals grouped by Smax ID
            - lig_conc_lst_per_id: List of ligand concentrations grouped by Smax ID
            - smax_guesses_unq: Smax guesses for each association signal
            - smax_guesses_shared: Smax guesses for each Smax ID
        """

        signals_steady_state = [np.median(assoc[-10:]) for assoc in self.assoc_lst]

        # Create a new list, that will contain one element per Smax ID
        # each element will be a list of steady-state signals

        self.signal_ss            = [] # convert to list of lists, one list per unique smax id
        self.lig_conc_lst_per_id  = [] # convert to list of lists, one list per unique smax id

        # Obtain the smax guesses, the maximum signal
        smax_guesses_unq    = [] # One element per association signal
        smax_guesses_shared = [] # One element per Smax ID

        for i,smax_id in enumerate(np.unique(self.smax_id)):

            idx = np.where(self.smax_id == smax_id)[0]

            self.signal_ss.append([signals_steady_state[i] for i in idx])
            self.lig_conc_lst_per_id.append([self.lig_conc_lst[i] for i in idx])

            smax = np.max(self.signal_ss[i])
            smax_guesses_unq.append(smax*1.5)
            smax_guesses_shared += [smax*1.5 for _ in range(len(idx))]

        self.smax_guesses_unq    = smax_guesses_unq
        self.smax_guesses_shared = smax_guesses_shared

        return None

    def fit_steady_state(self):

        """
        Fit the steady state signal using the one-site binding model.
        
        This function fits the steady state signal to a one-site binding model
        to estimate the equilibrium dissociation constant (Kd) and maximum signal (Smax).
        It also calculates the 95% confidence intervals for the Kd.
        
        Returns
        -------
        None
            Updates the following instance attributes:
            - params: The fitted parameters [Kd, Smax1, Smax2, ...]
            - signal_ss_fit: The fitted steady state signals
            - fit_params_ss: DataFrame with the fitted parameters
            - Kd_ss: The equilibrium dissociation constant
            - Smax_upper_bound_factor: Factor to determine the upper bound for Smax
        """

        self.clear_fittings()

        if self.signal_ss is None:

            self.get_steady_state()

        Kd_init = np.median(self.lig_conc_lst_per_id[0])
        p0      = [Kd_init] + [np.max(signal) for signal in self.signal_ss]

        kd_min  = np.min(self.lig_conc_lst_per_id[0]) / 1e3
        kd_max  = np.max(self.lig_conc_lst_per_id[0]) * 1e3

        # Find the upper bound for the Kd
        upper_bound = 1e3 if Kd_init >= 1 else 1e2

        low_bounds  = [kd_min]  + [x*0.5          for x in p0[1:]]
        high_bounds = [kd_max]  + [x*upper_bound  for x in p0[1:]]

        # testing - set upper bound for smax to smax
        high_bounds = [kd_max]  + [x*1  for x in p0[1:]]

        fit, cov, fit_vals = fit_steady_state_one_site(
            self.signal_ss,self.lig_conc_lst_per_id,
            p0,low_bounds,high_bounds)

        # Prepare the arguments for re-fitting fit_steady_state_one_site
        kwargs = {
            "signal_lst": self.signal_ss,
            "ligand_lst": self.lig_conc_lst_per_id
        }

        fit, cov, fit_vals, low_bounds, high_bounds = re_fit(
            fit=fit,
            cov=cov,
            fit_vals=fit_vals,
            fit_fx=fit_steady_state_one_site,
            low_bounds=low_bounds,
            high_bounds= high_bounds,
            times=3,
            **kwargs)

        self.Kd_ss   = fit[0]
        Smax         = fit[1:]

        self.params      = fit
        self.p0          = p0
        self.low_bounds  = low_bounds
        self.high_bounds = high_bounds

        self.signal_ss_fit = fit_vals

        n = np.sum([len(signal) for signal in self.signal_ss])
        p = len(p0)

        rss_desired = get_desired_rss(concat_signal_lst(self.signal_ss), concat_signal_lst(fit_vals), n, p)

        minKd, maxKd = steady_state_one_site_asymmetric_ci95(
            self.Kd_ss, self.signal_ss, self.lig_conc_lst_per_id, p0[1:],
            low_bounds[1:], high_bounds[1:], rss_desired)

        # Create a dataframe with the fitted parameters
        df_fit = pd.DataFrame({'Kd [µM]': self.Kd_ss,
                               'Kd_min95': minKd,
                               'Kd_max95': maxKd,
                               'Smax': Smax,
                               'Name': self.names})

        self.fit_params_ss = df_fit

        self.Smax_upper_bound_factor = get_smax_upper_bound_factor(self.Kd_ss)

        self.fit_params_ss = df_fit

        # Fit the steady state signal
        return None

    def get_k_off_initial_guess(self):
        """
        Get initial guess and bounds for k_off parameter.
        
        This method determines initial values and fitting bounds for Kd and k_off.
        If dissociation data is available, it fits the dissociation curves to get
        a better estimate of k_off.
        
        Returns
        -------
        tuple
            A tuple containing:
            - p0: Initial parameter guesses [Kd, k_off]
            - low_bounds: Lower bounds for the parameters
            - high_bounds: Upper bounds for the parameters
            
        Raises
        ------
        ValueError
            If Kd_ss is not set and fit_steady_state has not been run
        """

        # Initial guess for Kd_ss for single_cycle_kinetics
        if self.is_single_cycle:
            self.Kd_ss = np.median(self.lig_conc_lst)

        # Requires that Kd_ss is set
        if self.Kd_ss is None:
            raise ValueError("Kd_ss is not set. Please run fit_steady_state() first.")

        # We check if we have dissociation curves and fit them to get a better estimate of k_off
        if self.disso_lst is not None and self.time_disso_lst is not None:

            self.fit_one_site_dissociation()
            p0 = [self.Kd_ss, self.k_off]
            low_bounds  = [self.Kd_ss/7e2, self.k_off/7e2]
            high_bounds = [self.Kd_ss*7e2, self.k_off*7e2]

            self.clear_fittings()

        else:

            p0 = [self.Kd_ss,0.01]

            low_bounds  = [self.Kd_ss/7e2,1e-7]
            high_bounds = [self.Kd_ss*7e2,10]

        return p0, low_bounds, high_bounds

    def fit_one_site_association(self, shared_smax=True):
        """
        Fit the association curves using the one-site binding model.
        
        This function fits the association curves to a one-site binding model
        to estimate the equilibrium dissociation constant (Kd), dissociation
        rate constant (k_off), and maximum signal (Smax).
        
        Parameters
        ----------
        shared_smax : bool, optional
            Whether to share Smax across curves with the same Smax ID, default is True.
            
        Returns
        -------
        None
            Updates the following instance attributes:
            - params: The fitted parameters [Kd, k_off, Smax1, Smax2, ...]
            - signal_assoc_fit: The fitted association signals
            - Kd: The equilibrium dissociation constant
            - k_off: The dissociation rate constant
            - Smax: The maximum signal values
        """

        self.clear_fittings()

        p0, low_bounds, high_bounds = self.get_k_off_initial_guess()

        smax_guesses = self.smax_guesses_unq if shared_smax else self.smax_guesses_shared

        p0 += smax_guesses

        low_bounds  += [x/15 for x in p0[2:]]
        high_bounds += [x*self.Smax_upper_bound_factor for x in p0[2:]]

        self.p0           = p0
        self.low_bounds  = low_bounds
        self.high_bounds = high_bounds

        fit, cov, fit_vals = fit_one_site_association(
            self.assoc_lst,self.time_assoc_lst,self.lig_conc_lst,
            p0,low_bounds,high_bounds,smax_idx=self.smax_id,shared_smax=shared_smax
        )

        self.params = fit

        self.signal_assoc_fit = fit_vals

        self.Kd   = fit[0]
        self.k_off = fit[1]
        self.Smax = fit[2:]

        # Create a dataframe with the fitted parameters
        df_fit = pd.DataFrame({'Kd [µM]':   self.Kd,
                               'k_off [1/s]': self.k_off,
                               'Smax': self.Smax})

        error     = np.sqrt(np.diag(cov))
        rel_error = error/fit * 100

        df_error = pd.DataFrame({'Kd [µM]':   rel_error[0],
                                 'k_off': rel_error[1],
                                 'Smax': rel_error[2:]})

        self.fit_params_kinetics       = df_fit
        self.fit_params_kinetics_error = df_error

        return  None

    def fit_one_site_dissociation(self, time_limit=0):
        """
        Fit the dissociation curves using the one-site binding model.
        
        This function fits the dissociation curves to a one-site binding model
        to estimate the dissociation rate constant (k_off) and the initial signal (S0).
        
        Parameters
        ----------
        time_limit : float, optional
            Time limit for fitting (in seconds). If > 0, only fit data up to this time, 
            default is 0 (fit all data).
            
        Returns
        -------
        None
            Updates the following instance attributes:
            - params: The fitted parameters [k_off, S0_1, S0_2, ...]
            - signal_disso_fit: The fitted dissociation signals
            - k_off: The dissociation rate constant
            - fit_params_kinetics: DataFrame with the fitted parameters
            - fit_params_kinetics_error: DataFrame with the relative errors
        """

        self.clear_fittings()

        disso_lst     = self.disso_lst
        time_disso_lst = self.time_disso_lst

        # Fit only some data. If time_limit = 0, fit all data
        if time_limit > 0:

            disso_lst      = [x[t < (np.min(t)+time_limit)] for x,t in zip(disso_lst,time_disso_lst)]
            time_disso_lst = [t[t < (np.min(t)+time_limit)] for t in time_disso_lst]

        p0 = [0.1] + [np.max(signal) for signal in disso_lst]

        low_bounds  = [1e-7] + [x/5 for x in p0[1:]]
        high_bounds = [10]   + [x*5 for x in p0[1:]]

        self.p0           = p0
        self.low_bounds  = low_bounds
        self.high_bounds = high_bounds

        fit, cov, fit_vals = fit_one_site_dissociation(disso_lst,time_disso_lst,
                                                       p0,low_bounds,high_bounds)

        self.signal_disso_fit = fit_vals

        self.k_off = fit[0]

        # generate dataframe with the fitted parameters
        df_fit = pd.DataFrame({'k_off': self.k_off,
                               'S0': fit[1:]})

        self.fit_params_kinetics = df_fit

        # Add the fitted errors to the dataframe
        error     = np.sqrt(np.diag(cov))
        rel_error = error/fit * 100

        df_error = pd.DataFrame({'k_off [1/s]': rel_error[0],
                                 'S0':          rel_error[1:]})

        self.fit_params_kinetics_error = df_error
        self.params = fit

        return  None

    def fit_single_exponentials(self):
        """
        Fit single exponentials to the association signals.
        
        This method fits each association curve with a single exponential function
        and extracts the observed rate constants (k_obs).
        
        Returns
        -------
        None
            Updates the k_obs attribute with a list of observed rate constants
            for each association curve.
        """

        # Fit one exponential to each signal

        # Initialize a list to store the k_obs values - filld with NaNs
        k_obs = [np.nan for _ in range(len(self.assoc_lst))]

        i = 0
        for y,t in zip(self.assoc_lst,self.time_assoc_lst):

            fit_params, _, _ = fit_single_exponential(y,t)

            # Append the k_obs value to the list
            k_obs[i] = fit_params[2]

            i += 1

        self.k_obs = k_obs

        return None

    def fit_one_site_assoc_and_disso(self,shared_smax=True,fixed_t0=True,fit_ktr=False):

        self.clear_fittings()

        p0, low_bounds, high_bounds = self.get_k_off_initial_guess()

        # Extend the lower bound for Kd_ss
        low_bounds[0] = np.min([self.Kd_ss/1e3] + self.lig_conc_lst)  # Kd lower bound

        smax_guesses = self.smax_guesses_unq if shared_smax else self.smax_guesses_shared

        p0 += smax_guesses

        low_bounds  += [x/4 for x in p0[2:]]
        high_bounds += [x*self.Smax_upper_bound_factor for x in p0[2:]]

        smax_param_start = 2

        if fit_ktr:

            for i, _ in enumerate(self.smax_guesses_unq):
                p0.insert(2, 1e-4)
                low_bounds.insert(2, 1e-7)
                high_bounds.insert(2, 1)
                smax_param_start += 1

        if not fixed_t0:

            for i,_ in enumerate(self.smax_guesses_unq):
                p0.insert(2,0)
                low_bounds.insert(2,-0.01)
                high_bounds.insert(2,0.1)
                smax_param_start += 1

        if fit_ktr:

            fit, cov, fit_vals_assoc, fit_vals_disso = fit_one_site_assoc_and_disso_ktr(
                self.assoc_lst,self.time_assoc_lst,self.lig_conc_lst,
                self.disso_lst,self.time_disso_lst,
                p0,low_bounds,high_bounds,
                smax_idx=self.smax_id,
                shared_smax=shared_smax,
                fixed_t0=fixed_t0
            )

            kwargs = {
                'assoc_signal_lst' :self.assoc_lst,
                'assoc_time_lst' : self.time_assoc_lst,
                'analyte_conc_lst' : self.lig_conc_lst,
                'disso_signal_lst' :self.disso_lst,
                'disso_time_lst' : self.time_disso_lst,
                'shared_smax' : shared_smax,
                'smax_idx' : self.smax_id,
                'fixed_t0' : fixed_t0
            }

            fit, cov, fit_vals_assoc, fit_vals_disso,low_bounds, high_bounds = re_fit_2(
                fit=fit,
                cov=cov,
                fit_vals_a=fit_vals_assoc,
                fit_vals_b=fit_vals_disso,
                fit_fx=fit_one_site_assoc_and_disso_ktr,
                low_bounds=low_bounds,
                high_bounds=high_bounds,
                times=2,
                **kwargs
            )

        else:

            fit, cov, fit_vals_assoc, fit_vals_disso = fit_one_site_assoc_and_disso(
                self.assoc_lst,self.time_assoc_lst,self.lig_conc_lst,
                self.disso_lst,self.time_disso_lst,
                p0,low_bounds,high_bounds,
                smax_idx=self.smax_id,
                shared_smax=shared_smax,
                fixed_t0=fixed_t0)

        self.signal_assoc_fit = fit_vals_assoc
        self.signal_disso_fit = fit_vals_disso

        self.Kd   = fit[0]
        self.k_off = fit[1]
        self.Smax = fit[smax_param_start:]

        # Create a dataframe with the fitted parameters
        df_fit = pd.DataFrame({'Kd [µM]':     self.Kd,
                               'k_off [1/s]': self.k_off,
                               'Smax':        self.Smax})

        # Include the Kon, derived from the Kd and Koff
        df_fit['(Derived) k_on [1/µM/s]'] = df_fit['k_off [1/s]'] / df_fit['Kd [µM]']

        error     = np.sqrt(np.diag(cov))
        rel_error = error/fit * 100

        df_error = pd.DataFrame({'Kd [µM]':   rel_error[0],
                                 'k_off [1/s]': rel_error[1],
                                 'Smax': rel_error[smax_param_start:]})

        # Add the t0 parameter
        if not fixed_t0:

            t0       = fit[3:3+len(np.unique(self.smax_id))]
            t0_error = rel_error[3:3+len(np.unique(self.smax_id))]

            if not shared_smax:

                t0_all       = expand_parameter_list(t0,       self.smax_id)
                t0_error_all = expand_parameter_list(t0_error, self.smax_id)

                df_fit['t0']   = t0_all
                df_error['t0'] = t0_error_all

            else:

                df_fit['t0']   = t0
                df_error['t0'] = t0_error

        if fit_ktr:

            n_ktr      = len(np.unique(self.smax_id))
            idx_start  = 2+(not fixed_t0)*n_ktr
            ktrs       = fit[idx_start:smax_param_start]
            ktrs_error = rel_error[idx_start:smax_param_start]

            if not shared_smax:

                ktr_all       = expand_parameter_list(ktrs,       self.smax_id)
                ktr_error_all = expand_parameter_list(ktrs_error, self.smax_id)

                df_fit['Ktr']   = ktr_all
                df_error['Ktr'] = ktr_error_all

            else:

                df_fit['Ktr']   = ktrs
                df_error['Ktr'] = ktrs_error

        self.fit_params_kinetics       = df_fit
        self.fit_params_kinetics_error = df_error

        self.params      = fit
        self.p0          = p0
        self.low_bounds  = low_bounds
        self.high_bounds = high_bounds

        return  None

    def calculate_ci95(self, shared_smax=True, fixed_t0=True, fit_ktr=False):
        """
        Calculate 95% confidence intervals for the fitted parameters.
        
        This method computes asymmetrical 95% confidence intervals for the
        equilibrium dissociation constant (Kd) and the dissociation rate constant (k_off).
        
        Parameters
        ----------
        shared_smax : bool, optional
            Whether Smax was shared across curves with the same Smax ID, default is True.
        fixed_t0 : bool, optional
            Whether the time offset (t0) was fixed to zero, default is True.
        fit_ktr : bool, optional
            Whether mass transport rate constants were included in the model, default is False.
            
        Returns
        -------
        None
            Updates the fit_params_kinetics_ci95 attribute with a DataFrame containing
            the 95% confidence intervals for Kd and k_off.
            
        Notes
        -----
        This method will create an empty DataFrame if the confidence interval calculation fails.
        """

        try:

        # Compute asymmetrical 95% confidence intervals
            if not fit_ktr:

                exp_signal_concat = concat_signal_lst(self.assoc_lst+self.disso_lst)
                fit_signal_concat = concat_signal_lst(self.signal_assoc_fit+self.signal_disso_fit)
                n                 = len(exp_signal_concat)
                p                 = len(self.p0)

                rss_desired = get_desired_rss(exp_signal_concat,
                                              fit_signal_concat,
                                              n, p)

                Kd_min, Kd_max = one_site_assoc_and_disso_asymmetric_ci95(
                    self.Kd,rss_desired,
                    self.assoc_lst,self.time_assoc_lst,
                    self.lig_conc_lst,
                    self.disso_lst,self.time_disso_lst,
                    self.p0[1:],self.low_bounds[1:],self.high_bounds[1:],
                    self.smax_id,shared_smax=shared_smax,fixed_t0=fixed_t0)

                p0 = self.p0[:1] + self.p0[2:]
                low_bounds = self.low_bounds[:1] + self.low_bounds[2:]
                high_bounds = self.high_bounds[:1] + self.high_bounds[2:]

                koff_min, koff_max  = one_site_assoc_and_disso_asymmetric_ci95_koff(
                    self.k_off,rss_desired,
                    self.assoc_lst,self.time_assoc_lst,
                    self.lig_conc_lst,
                    self.disso_lst,self.time_disso_lst,
                    p0,low_bounds,high_bounds,
                    self.smax_id,shared_smax=shared_smax,fixed_t0=fixed_t0)

                # Convert ci95_Kd and ci95_koff to a nice Table
                header = ['Parameter','95% CI lower','95% CI upper']
                row1   = ['Kd [µM]',Kd_min,Kd_max]
                row2   = ['k_off [1/s]',koff_min,koff_max]

                df = pd.DataFrame([row1,row2],columns=header)

                self.fit_params_kinetics_ci95 = df

        except:

            # Generate an empty dataframe if the procedure did not work
            self.fit_params_kinetics_ci95 = pd.DataFrame()

        return  None

    def fit_one_site_if_assoc_and_disso(self, shared_smax=True):
        """
        Fit the association and dissociation signals using the induced fit model.
        
        This method first fits a simple one-site model to get initial parameter estimates,
        then explores different values of the conformational change rate constants (kc and krev)
        to find good initial guesses for the induced fit model. Finally, it fits the full
        induced fit model with these initial guesses.
        
        Parameters
        ----------
        shared_smax : bool, optional
            Whether to share Smax across curves with the same Smax ID, default is True.
            
        Returns
        -------
        None
            Updates the following instance attributes:
            - params: The fitted parameters [k_on, k_off, k_c, k_rev, Smax1, Smax2, ...]
            - signal_assoc_fit: The fitted association signals
            - signal_disso_fit: The fitted dissociation signals
            - k_on: The association rate constant
            - k_off: The dissociation rate constant
            - k_c: The forward conformational change rate constant
            - k_rev: The reverse conformational change rate constant
            - Smax: The maximum signal values
            - fit_params_kinetics: DataFrame with the fitted parameters
        """

        # Fit first a model without induced-fit

        self.fit_one_site_assoc_and_disso(shared_smax=shared_smax)

        # Get the initial parameters from the single site fit
        p0          = self.p0
        low_bounds  = self.low_bounds
        high_bounds = self.high_bounds

        # Find kon from Kd (1st fitted param) and koff (2nd fitted param)
        kon = p0[1] / p0[0]

        p0[0]          = kon
        low_bounds[0]  = kon / 1e3
        high_bounds[0] = kon * 1e3

        # We need to find good initial guesses for the kc and krev parameters
        kc_init    = np.logspace(-4, 1, 6)
        k_rev_init = kc_init

        # Build combinations as numpy array, filter kc >= krev/100
        combinations = np.array(list(itertools.product(kc_init, k_rev_init)))
        mask = combinations[:, 0] >= combinations[:, 1] / 100
        combinations = combinations[mask]

        rss_init    = np.inf
        best_kc     = None
        best_krev   = None
        best_params = None

        # We create a subsample of the time points to speed up the fitting for the initial guess
        time_assoc_lst_subsampled = [subset_data(t) for t in self.time_assoc_lst]
        time_disso_lst_subsampled = [subset_data(t) for t in self.time_disso_lst]
        assoc_lst_subsampled      = [subset_data(y) for y in self.assoc_lst]
        disso_lst_subsampled      = [subset_data(y) for y in self.disso_lst]

        # Loop through each combination of kc and krev
        # Apply fit_induced_fit_sites_assoc_and_disso with fixed kon2 and koff2 (corresponding to kc and krev)
        for kc, krev in combinations:

            try:

                params, cov, fit_vals_assoc, fit_vals_disso = fit_induced_fit_sites_assoc_and_disso(
                    assoc_lst_subsampled,time_assoc_lst_subsampled,self.lig_conc_lst,
                    disso_lst_subsampled,time_disso_lst_subsampled,
                    p0,low_bounds,high_bounds,
                    smax_idx=self.smax_id,
                    shared_smax=shared_smax,
                    fixed_t0=True,
                    fixed_kon2 = True, # k_on2 refers here to the forward rate constant of the induced fit
                    kon2_value=kc,
                    fixed_koff2 = True, # k_off2 refers here to the reverse rate constant of the induced fit
                    koff2_value=krev
                )

                # Calculate the residuals for the signals - use the subsampled data
                rss_asso  = np.sum([np.sum((y - fit_y)**2) for y, fit_y in zip(assoc_lst_subsampled, fit_vals_assoc)])
                rss_disso = np.sum([np.sum((y - fit_y)**2) for y, fit_y in zip(disso_lst_subsampled, fit_vals_disso)])

                rss = rss_asso + rss_disso

                if rss < rss_init:

                    rss_init = rss
                    best_kc  = kc
                    best_krev = krev
                    best_params = params

            except:

                # If the fit fails, just continue to the next combination
                continue

        # Insert the kc and krev parameters into the p0, low_bounds and high_bounds lists

        factor = 1e3  # Factor to scale the bounds for kc and krev

        p0.insert(2, best_kc)
        low_bounds.insert(2, best_kc/factor)
        high_bounds.insert(2, best_kc*factor)
        p0.insert(3, best_krev)
        low_bounds.insert(3, best_krev/factor)
        high_bounds.insert(3, best_krev*factor)

        # Replace the kon and koff with the best parameters
        p0[0]  = best_params[0]  # kon
        p0[1]  = best_params[1]  # koff

        low_bounds[0]  = best_params[0] / factor  # kon
        high_bounds[0] = best_params[0] * factor

        low_bounds[1]  = best_params[1] / factor  # koff
        high_bounds[1] = best_params[1] * factor

        fit, cov, fit_vals_assoc, fit_vals_disso = fit_induced_fit_sites_assoc_and_disso(
            self.assoc_lst,self.time_assoc_lst,self.lig_conc_lst,
            self.disso_lst,self.time_disso_lst,
            p0,low_bounds,high_bounds,
            smax_idx=self.smax_id,
            shared_smax=shared_smax,
            fixed_t0=True)

        self.signal_assoc_fit = fit_vals_assoc
        self.signal_disso_fit = fit_vals_disso

        self.k_on, self.k_off, self.k_c, self.k_rev, *self.Smax = fit

        # Create a dataframe with the fitted parameters
        df_fit = pd.DataFrame({'k_on [1/(s*µM)]':  self.k_on,
                               'k_off [1/s]':      self.k_off,
                               'k_c [1/s]':        self.k_c,
                               'k_rev [1/s]':      self.k_rev,
                               'Smax':             self.Smax})

        self.fit_params_kinetics = df_fit

        self.params      = fit
        self.p0          = p0
        self.low_bounds  = low_bounds
        self.high_bounds = high_bounds

        return None

    def create_export_df(self,type='raw'):

        """
        Create a data frame with the raw or fitted signals for the association and dissociation phases.

        Parameters
        ----------
        type : str, optional
         The type of signal to include in the 'Signal' column. If 'raw', the raw association and dissociation signals are used.
         If 'fit', the fitted association and dissociation signals are used. Default is 'raw'.

        Returns
        -------
        pd.DataFrame
            A DataFrame with the following columns:
            - 'Time': Time points for the association and dissociation signals.
            - 'Signal': The corresponding signal values (raw or fitted).
            - 'Analyte_concentration_micromolar': The analyte concentration in micromolar for each signal.
            - 'Type': A categorical variable indicating whether the signal is from the association or dissociation phase.
            - 'ID': The identifier corresponding to the fitter name for each signal.

        """

        df_list = []

        for i, name in enumerate(self.names):

            time_assoc = self.time_assoc_lst[i]

            if type == 'raw':
                signal_assoc = self.assoc_lst[i]
            else:
                signal_assoc = self.signal_assoc_fit[i]

            lig_conc = self.lig_conc_lst[i]

            dic = {
                'Time': time_assoc,
                'Signal': signal_assoc,
                'Analyte_concentration_micromolar': lig_conc,
                'Type': 'Association',
                'ID': name
            }

            df = pd.DataFrame(dic)
            df_list.append(df)

            if self.disso_lst is not None and self.time_disso_lst is not None:

                time_disso = self.time_disso_lst[i]

                signal_disso = None

                if type == 'raw':
                    signal_disso = self.disso_lst[i]
                else:

                    if self.signal_disso_fit is not None:

                        signal_disso = self.signal_disso_fit[i]

                # Append only if we have signal_disso
                if signal_disso is not None:

                    dic = {
                        'Time': time_disso,
                        'Signal': signal_disso,
                        'Analyte_concentration_micromolar': 0,
                        'Type': 'Dissociation',
                        'ID': name
                    }

                    df = pd.DataFrame(dic)
                    df_list.append(df)

        df = pd.concat(df_list, ignore_index=True)

        return df




from .main              import KineticsAnalyzer
from .fitter_solution   import KineticsFitterSolution
from .fitter_surface    import KineticsFitter
from .octet             import OctetExperiment
from .gator             import GatorExperiment
from .kingenie_solution import KinGenieCsvSolution
from .kingenie_surface  import KinGenieCsv

from .utils.processing import (
    guess_experiment_name,
    guess_experiment_type,
    concat_signal_lst,
    get_palette,
    get_plotting_df,
    subset_data,
    sample_type_to_letter,
    get_colors_from_numeric_values
)

from .utils.math import (
    single_exponential,
    double_exponential,
    median_filter,
    rss_p,
    get_rss,
    get_desired_rss
)

from .utils.signal_surface import (
    steady_state_one_site,
    one_site_association_analytical,
    one_site_dissociation_analytical,
    ode_one_site_mass_transport_association,
    ode_one_site_mass_transport_dissociation,
    solve_ode_one_site_mass_transport_association,
    solve_ode_one_site_mass_transport_dissociation,
    differential_matrix_association_induced_fit,
    differential_matrix_association_conformational_selection,
    differential_matrix_dissociation_induced_fit,
    differential_matrix_dissociation_conformational_selection,
    constant_vector_induced_fit,
    constant_vector_conformational_selection,
    solve_all_states_fast,
    solve_steady_state,
    solve_induced_fit_association,
    solve_conformational_selection_association,
    solve_induced_fit_dissociation,
    solve_conformational_selection_dissociation,
    ode_mixture_analyte_association,
    solve_ode_mixture_analyte_association,
    ode_mixture_analyte_dissociation,
    solve_ode_mixture_analyte_dissociation
)

from .utils.signal_solution  import (
    solve_ode_one_site_insolution,
    signal_ode_one_site_insolution,
    ode_induced_fit_insolution,
    solve_ode_induced_fit_insolution,
    signal_ode_induced_fit_insolution,
    ode_conformational_selection_insolution,
    solve_ode_conformational_selection_insolution,
    signal_ode_conformational_selection_insolution,
    get_initial_concentration_conformational_selection,
    get_kobs_induced_fit,
    get_kobs_conformational_selection
)

from .utils.fitting_surface  import (
    guess_initial_signal,
    fit_steady_state_one_site,
    steady_state_one_site_asymmetric_ci95,
    fit_one_site_association,
    fit_one_site_dissociation,
    fit_one_site_assoc_and_disso,
    fit_induced_fit_sites_assoc_and_disso,
    fit_one_site_assoc_and_disso_ktr,
    one_site_assoc_and_disso_asymmetric_ci95,
    one_site_assoc_and_disso_asymmetric_ci95_koff
)

from .utils.fitting_general  import (
    fit_single_exponential,
    fit_double_exponential
)

from .utils.fitting_solution import (
    fit_one_site_solution,
    fit_induced_fit_solution,
    find_initial_parameters_induced_fit_solution
)

from .utils.plotting import (
    config_fig,
    plot_plate_info,
    plot_traces,
    plot_traces_all,
    plot_steady_state,
    plot_association_dissociation
)

# Package metadata
__version__ = '0.1.0'
__author__ = 'osvalB'
__email__ = 'oburastero@gmail.com'
__description__ = 'A Python package for kinetics data analysis'


"""
Color palettes for plotting.

This module provides color palettes that can be used for visualizing data
in plots. These palettes are collections of colors represented as hex values.
"""

__all__ = [
    "VIRIDIS",
    "set1_colors",
    "set3_colors"
]

VIRIDIS = [
    "#440154", "#471063", "#481d6f", "#472a7a", "#414487", "#3c4f8a", "#375a8c",
    "#32648e", "#2a788e", "#26828e", "#228b8d", "#1f958b", "#22a884", "#2cb17e",
    "#3bbb75", "#4ec36b", "#7ad151", "#95d840", "#b0dd2f", "#cae11f", "#fde725"
]
"""
Viridis color palette.

A perceptually uniform color map that is readable by those with colorblindness.
Contains 21 hex color values transitioning from dark purple to yellow.
"""

set1_colors = [
    "#e41a1c",  # red
    "#377eb8",  # blue
    "#4daf4a",  # green
    "#984ea3",  # purple
    "#ff7f00",  # orange
    "#ffff33",  # yellow
    "#a65628",  # brown
    "#f781bf",  # pink
    "#999999"   # gray
]
"""
Set1 color palette.

A qualitative color palette suitable for categorical data.
Contains 9 distinct colors that are visually distinguishable.
"""

set3_colors = [
    "#8dd3c7",  # teal
    "#ffffb3",  # light yellow
    "#bebada",  # lavender
    "#fb8072",  # salmon
    "#80b1d3",  # light blue
    "#fdb462",  # peach
    "#b3de69",  # light green
    "#fccde5",  # pink
    "#d9d9d9",  # gray
    "#bc80bd",  # purple
    "#ccebc5",  # mint
    "#ffed6f"   # yellow
]
"""
Set3 color palette.

A qualitative color palette with softer, pastel-like colors.
Contains 12 distinct colors suitable for categorical data with many categories.
"""


import itertools
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit

from .math import (
    single_exponential,
    double_exponential
)

__all__ = ['fit_single_exponential', 'fit_double_exponential','re_fit','re_fit_2']

def convert_to_numpy_array(data):
    """
    Convert input data to a numpy array if it is not already one.

    Parameters
    ----------
    data : list or np.ndarray
        Input data to be converted.

    Returns
    -------
    np.ndarray
        Converted numpy array.
    """
    if not isinstance(data, np.ndarray):
        return np.array(data)

    return data

def fit_single_exponential(y, t, min_log_k=-5, max_log_k=5, log_k_points=50):
    """
    Fit a single exponential to a signal.

    Parameters
    ----------
    y : np.ndarray
        Signal values.
    t : np.ndarray
        Time values.
    min_log_k : float, optional
        Minimum logarithmic value for k_obs initial estimation, default is -5.
    max_log_k : float, optional
        Maximum logarithmic value for k_obs initial estimation, default is 5.
    log_k_points : int, optional
        Number of points in the logarithmic scale for k_obs initial estimation, default is 50.

    Returns
    -------
    np.ndarray
        Fitted parameters (a0, a1, k_obs).
    np.ndarray
        Covariance matrix of the fitted parameters.
    np.ndarray
        Fitted y values.
    """
    # Convert to numpy array, if needed,
    y = convert_to_numpy_array(y)
    t = convert_to_numpy_array(t)

    t = t - np.min(t)  # Start at zero

    possible_k = np.logspace(min_log_k, max_log_k, log_k_points)

    # define a single exponential function reduced to fit only a0 and a1
    def single_exponential_reduced(t, a0, a1):
        return a0 + a1 * np.exp(-k_obs * t)

    rss         = np.inf
    best_params = None
    best_k_obs  = None

    # Loop through each k_obs value
    for k_obs in possible_k:

        # Fit the reduced model
        params, cov = curve_fit(single_exponential_reduced, t, y, p0=[0, np.max(y)])

        # Calculate the residual sum of squares
        rss_temp = np.sum((y - single_exponential_reduced(t, *params)) ** 2)

        if rss_temp < rss:
            rss         = rss_temp
            best_params = params
            best_k_obs  = k_obs

    p0 = [best_params[0], best_params[1], best_k_obs]

    fit_params, cov = curve_fit(single_exponential, t, y, p0=p0)

    fit_y = single_exponential(t, *fit_params)

    return fit_params, cov, fit_y

def fit_double_exponential(y, t, min_log_k=-4, max_log_k=4, log_k_points=24):
    """
    Fit a double exponential to a signal.

    Parameters
    ----------
    y : np.ndarray
        Signal values.
    t : np.ndarray
        Time values.
    min_log_k : float, optional
        Minimum logarithmic value for k_obs initial estimation, default is -4.
    max_log_k : float, optional
        Maximum logarithmic value for k_obs initial estimation, default is 4.
    log_k_points : int, optional
        Number of points in the logarithmic scale for k_obs initial estimation, default is 22.

    Returns
    -------
    np.ndarray
        Fitted parameters (a0, a1, k_obs1, a2, k_obs2).
    np.ndarray
        Covariance matrix of the fitted parameters.
    np.ndarray
        Fitted y values.
    """
    # Convert to numpy array, if needed,
    y = convert_to_numpy_array(y)
    t = convert_to_numpy_array(t)

    # Define a two columns dataframe with possible values for k obs 1 and k obs 2
    # Evenly spaced in logarithmic scale
    k_obs_1 = np.logspace(min_log_k, max_log_k, log_k_points)

    combinations = np.array(list(itertools.product(k_obs_1, k_obs_1)))

    combinations_df = pd.DataFrame(combinations, columns=['k_obs_1', 'k_obs_2'])

    # Remove all combinations where k_obs_1 is larger than k_obs_2
    combinations_df = combinations_df[combinations_df['k_obs_1'] < combinations_df['k_obs_2']]

    rss_min = np.inf

    # Loop through each combination of k_obs_1 and k_obs_2
    for index, row in combinations_df.iterrows():

        k_obs1 = row['k_obs_1']
        k_obs2 = row['k_obs_2']

        def double_exponential_reduced(t, a0, a1, a2):

            return a0 + a1 * np.exp(-k_obs1 * t) + a2 * np.exp(-k_obs2 * t)

        try:

            params, cov = curve_fit(double_exponential_reduced, t, y, p0=[0, np.max(y), np.max(y)])

            rss = np.sum((y - double_exponential_reduced(t, *params)) ** 2)

            if rss < rss_min:
                rss_min = rss
                best_params = params
                best_k_obs1 = k_obs1
                best_k_obs2 = k_obs2

        except:

            pass

    # Now fit the full double exponential model
    a0 = best_params[0]
    a1 = best_params[1]
    a2 = best_params[2]

    p0 = [a0, a1, best_k_obs1, a2, best_k_obs2]

    params, cov = curve_fit(double_exponential, t, y, p0=p0)
    fitted_y    = double_exponential(t, *params)

    return params, cov, fitted_y

def fit_many_double_exponential(signal_lst, time_lst, min_log_k=-4, max_log_k=4, log_k_points=22):
    """
    Fit a double exponential to many signals.
    
    Parameters
    ----------
    signal_lst : list of np.ndarray
        List of signals to fit.
    time_lst : list of np.ndarray
        List of time arrays corresponding to each signal.
    min_log_k : float, optional
        Minimum logarithmic value for k_obs initial estimation, default is -4.
    max_log_k : float, optional
        Maximum logarithmic value for k_obs initial estimation, default is 4.
    log_k_points : int, optional
        Number of points in the logarithmic scale for k_obs initial estimation, default is 22.
        
    Returns
    -------
    list
        List of slowest k_obs values for each signal.
    list
        List of second slowest k_obs values for each signal.
    list
        List of fitted values for each signal.
    list
        List of the relative errors for the slowest k_obs values.
    list
        List of the relative errors for the second slowest k_obs values.
        
    Notes
    -----
    This function iterates over each signal and time pair, fits a double exponential model,
    and returns the slowest and second slowest k_obs values along with the fitted values.
    If fitting fails for a signal, it will skip that signal and return NaN for k_obs values 
    and None for fitted values. The function assumes that the input signals and time arrays 
    are of the same length.
    """

    k_obs_1 = [np.nan for _ in range(len(signal_lst))]
    k_obs_2 = [np.nan for _ in range(len(signal_lst))]
    y_pred =  [None   for _ in range(len(signal_lst))]

    k_obs_1_err = [np.nan for _ in range(len(signal_lst))]
    k_obs_2_err = [np.nan for _ in range(len(signal_lst))]

    i = 0
    for y, t in zip(signal_lst, time_lst):

        try:

            params, cov, fitted_y = fit_double_exponential(y, t, min_log_k=min_log_k, max_log_k=max_log_k,
                                                           log_k_points=log_k_points)

            errors = (np.sqrt(cov) / params) * 100

            k_obs_both = [params[2], params[4]]
            k_obs_both_err = [errors[2], errors[4]]

            # Find the index of the slowest k_obs
            min_idx = np.argmin(k_obs_both)
            max_idx = 1 - min_idx

            slowest_k = k_obs_both[min_idx]
            second_k = k_obs_both[max_idx]

            slowest_k_err = k_obs_both_err[min_idx]
            second_k_err = k_obs_both_err[max_idx]

            k_obs_1[i] = slowest_k
            k_obs_2[i] = second_k
            y_pred[i] = fitted_y

            k_obs_1_err[i] = slowest_k_err
            k_obs_2_err[i] = second_k_err

        except:

            pass

        i += 1

    return k_obs_1, k_obs_2, y_pred, k_obs_1_err, k_obs_2_err

def expand_high_bound(value, factor=10):
    """
    Multiply/divide the value by a factor and return the new value.
    
    If the parameter is negative, we divide by the factor, otherwise we multiply.
    
    Parameters
    ----------
    value : float
        Parameter to expand.
    factor : float, optional
        Factor to multiply/divide by, default is 10.
        
    Returns
    -------
    float
        New parameter value after expansion.
    """
    new_value = value * factor if value >= 0 else value / factor

    return new_value

def expand_low_bound(value, factor=10):
    """
    Multiply/divide the value by a factor and return the new value.
    
    If the parameter is negative, we multiply by the factor, otherwise we divide.
    
    Parameters
    ----------
    value : float
        Parameter to expand.
    factor : float, optional
        Factor to multiply/divide by, default is 10.
        
    Returns
    -------
    float
        New parameter value after expansion.
    """
    new_value = value * factor if value < 0 else value / factor

    return new_value

def re_fit(fit, cov, fit_vals, fit_fx, low_bounds, high_bounds, times, **kwargs):
    """
    Re-fit data by evaluating the difference between fitted parameters and bounds.
    
    If the difference is less than 2 percent, the bounds are relaxed by a factor of 10
    and the fitting is repeated.
    
    Parameters
    ----------
    fit : list
        Fitted parameters.
    cov : np.ndarray
        Covariance matrix of the fitted parameters.
    fit_vals : list
        Fitted values for each signal, same dimensions as signal_lst.
    fit_fx : callable
        Function to fit the data - returns fit, cov, fit_vals.
    low_bounds : list
        Lower bounds for the parameters.
    high_bounds : list
        Upper bounds for the parameters.
    times : int
        Number of times to re-fit the data.
    **kwargs
        Additional arguments to pass to the fitting function.
        
    Returns
    -------
    np.ndarray
        Fitted parameters after re-fitting.
    np.ndarray
        Covariance matrix of the fitted parameters after re-fitting.
    list
        Fitted values for each signal after re-fitting.
    np.ndarray
        Lower bounds for the parameters after re-fitting.
    np.ndarray
        Upper bounds for the parameters after re-fitting.
    """
    fit         = np.array(fit).astype(float)
    low_bounds  = np.array(low_bounds).astype(float)
    high_bounds = np.array(high_bounds).astype(float)

    for _ in range(times):

        # upper bounds may contain np.inf, which will cause a division error
        difference_to_upper = np.abs(np.array([(a-b)/a if a != np.inf else np.inf for a, b in zip(high_bounds, fit)]))
        difference_to_lower = np.abs( (fit - low_bounds) / fit )

        c1 = any(difference_to_upper < 0.02)
        c2 = any(difference_to_lower < 0.02)

        if c1:

            # Relax bounds by a factor of 10 - only those that are too close to the upper bound
            high_bounds = [expand_high_bound(value, factor=10) if diff < 0.02 else value for value, diff in zip(high_bounds, difference_to_upper)]
            high_bounds = np.array(high_bounds)

        if c2:

            low_bounds = [expand_low_bound(value, factor=10) if diff < 0.02 else value for value, diff in zip(low_bounds, difference_to_lower)]
            low_bounds = np.array(low_bounds)

        if c1 or c2:

           # Some fitting functions return 3 elements: fit, cov, fit_vals, while others return those 3 and the parameter names
            result = fit_fx(
                initial_parameters=fit,
                low_bounds=low_bounds,
                high_bounds=high_bounds,
                **kwargs)

            if len(result) == 4:
                fit, cov, fit_vals, param_names = result
            else:
                fit, cov, fit_vals = result

    return fit, cov, fit_vals, low_bounds, high_bounds

def re_fit_2(fit, cov, fit_vals_a, fit_vals_b, fit_fx, low_bounds, high_bounds, times, **kwargs):
    """
    Re-fit data for two signals by evaluating the difference between fitted parameters and bounds.
    
    If the difference is less than 2 percent, the bounds are relaxed by a factor of 10
    and the fitting is repeated. This function is used for fitting two signals at once,
    e.g. fitting association and dissociation signals simultaneously.
    
    Parameters
    ----------
    fit : list
        Fitted parameters.
    cov : np.ndarray
        Covariance matrix of the fitted parameters.
    fit_vals_a : list
        Fitted values for the first set of signals.
    fit_vals_b : list
        Fitted values for the second set of signals.
    fit_fx : callable
        Function to fit the data - returns fit, cov, fit_vals_a, fit_vals_b.
    low_bounds : list
        Lower bounds for the parameters.
    high_bounds : list
        Upper bounds for the parameters.
    times : int
        Number of times to re-fit the data.
    **kwargs
        Additional arguments to pass to the fitting function.
        
    Returns
    -------
    np.ndarray
        Fitted parameters after re-fitting.
    np.ndarray
        Covariance matrix of the fitted parameters after re-fitting.
    list
        Fitted values for the first set of signals after re-fitting.
    list
        Fitted values for the second set of signals after re-fitting.
    np.ndarray
        Lower bounds for the parameters after re-fitting.
    np.ndarray
        Upper bounds for the parameters after re-fitting.
    """
    fit         = np.array(fit).astype(float)
    low_bounds  = np.array(low_bounds).astype(float)
    high_bounds = np.array(high_bounds).astype(float)

    for _ in range(times):

        difference_to_upper = np.abs(np.array([(a-b)/a if a != np.inf else np.inf for a, b in zip(high_bounds, fit)]))
        difference_to_lower = np.abs( (fit - low_bounds) / fit)

        c1 = any(difference_to_upper < 0.02)
        c2 = any(difference_to_lower < 0.02)

        if c1:

            # Relax bounds by a factor of 10 - only those that are too close to the upper bound
            high_bounds = [expand_high_bound(value, factor=10) if diff < 0.02 else value for value, diff in zip(high_bounds, difference_to_upper)]
            high_bounds = np.array(high_bounds)

        if c2:

            low_bounds = [expand_low_bound(value, factor=10) if diff < 0.02 else value for value, diff in zip(low_bounds, difference_to_lower)]
            low_bounds = np.array(low_bounds)

        if c1 or c2:

            fit, cov, fit_vals_a, fit_vals_b  = fit_fx(
                initial_parameters=fit,
                low_bounds=low_bounds,
                high_bounds=high_bounds,
                **kwargs)

    return fit, cov, fit_vals_a, fit_vals_b, low_bounds, high_bounds



class SolutionBasedExp:
    """
    Base class for solution-based kinetics experiments.
    
    This class serves as a foundation for classes that handle solution-based
    kinetic experiments data.
    
    Parameters
    ----------
    name : str
        Name of the experiment.
    type : str
        Type of the experiment.
        
    Attributes
    ----------
    name : str
        Name of the experiment.
    type : str
        Type of the experiment.
    xs : list of numpy.ndarray or None
        List of x values (time) for each trace.
    ys : list of numpy.ndarray or None
        List of y values (signal) for each trace.
    no_traces : int or None
        Number of traces.
    traces_names : list of str or None
        List of trace names.
    traces_names_unique : list of str or None
        List of unique trace names (includes experiment name).
    conc_df : pandas.DataFrame or None
        DataFrame with concentration information.
    traces_loaded : bool
        Whether traces have been loaded.
    """

    def __init__(self, name, type):
        """Initialize the SolutionBasedExp instance.
        
        Parameters
        ----------
        name : str
            Name of the experiment.
        type : str
            Type of the experiment.
        """

        self.name                = name
        self.type                = type

        self.xs                  = None
        self.ys                  = None
        self.no_traces           = None
        self.traces_names        = None
        self.traces_names_unique = None
        self.conc_df             = None
        self.traces_loaded       = False

    def create_unique_traces_names(self):
        """
        Create unique trace names by appending the experiment name to the trace names.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
            Updates the traces_names_unique attribute.
        
        Notes
        -----
        This method creates unique identifiers for each trace by combining
        the experiment name with the individual trace names.
        """

        self.traces_names_unique = [self.name + ' ' + trace_name for trace_name in self.traces_names]

        return None

    def cut_off_time(self,time):

        """
        Remove all data before a specified time.

        Parameters
        ----------
        time : float
            The time to cut off the data.
        """

        for i in range(len(self.xs)):

            self.ys[i] = self.ys[i][self.xs[i] <= time]
            self.xs[i] = self.xs[i][self.xs[i] <= time]

        return None

    def get_trace_xy(self, trace_name, type='y'):
        """
        Return the x or y values of a specific trace.
        
        Parameters
        ----------
        trace_name : str
            Name of the trace to retrieve data from.
        type : {'x', 'y'}, optional
            Type of data to return. Default is 'y'.
            
        Returns
        -------
        numpy.ndarray
            The x or y values of the specified trace.
            
        Notes
        -----
        This method finds the trace by name in the traces_names list
        and returns either its x values (time) or y values (signal).
        """

        trace_id = self.traces_names.index(trace_name)

        if type == 'x':

            return self.xs[trace_id]

        else:

            return self.ys[trace_id]




import pytest
import numpy as np

from pykingenie.utils.signal_surface import (
    steady_state_one_site,
    one_site_association_analytical,
    one_site_dissociation_analytical,
    solve_ode_one_site_mass_transport_association,
    solve_ode_one_site_mass_transport_dissociation,
    solve_induced_fit_association,
    solve_induced_fit_dissociation,
    solve_conformational_selection_association,
    solve_conformational_selection_dissociation,
    solve_ode_mixture_analyte_association,
    solve_ode_mixture_analyte_dissociation
)

def test_steady_state_one_site():

    C = np.logspace(-9, -3, 20)  # Concentration range from 10^-9 to 10^-3
    Kd = 1e-6  # Dissociation constant
    R_max = 1.0  # Maximum response

    response = steady_state_one_site(C, R_max, Kd)

    # Check if the response is a numpy array
    assert isinstance(response, np.ndarray), "The response should be a numpy array."

    # Check that the last value is close to R_max
    assert np.isclose(response[-1], R_max, atol=1e-3), "The last value of the response should be close to R_max."

    # Check that the response is non-negative
    assert np.all(response >= 0), "The response should be non-negative."

    # Check that the response increases with concentration
    assert np.all(np.diff(response) >= 0), "The response should be non-decreasing with increasing concentration."

def test_one_site_association_analytical():

    t = np.linspace(0, 10, 10)
    s0 = 0 # initial signal
    s_max = 1.0 # maximum signal
    k_off = 0.05 # dissociation rate constant
    Kd = 0.5 # equilibrium dissociation constant
    A = 1.0 # total concentration of A
    t0=0

    result = one_site_association_analytical(t, s0, s_max, k_off, Kd, A, t0)

    # Check if the result is a numpy array
    assert isinstance(result, np.ndarray), "The result should be a numpy array."

    # Check the length of the result
    assert len(result) == len(t), "The length of the result should match the length of t."

    # Check the first value of the result
    assert np.isclose(result[0], s0), "The first value of the result should match the initial signal."

    # Check increasing behavior
    assert np.all(np.diff(result) >= 0), "The result should be non-decreasing with increasing time."

    # Assuming infinite concentration of A, the last value should be close to s_max
    result = one_site_association_analytical(t, s0, s_max, k_off, Kd, 1e6, t0)

    assert np.isclose(result[-1], s_max), "The last value of the result should match the expected value based on the maximum signal and dissociation"

def test_one_site_dissociation_analytical():
    t = np.linspace(0, 10, 10)
    s0 = 1.0 # initial signal
    k_off = 0.05 # dissociation rate constant

    result = one_site_dissociation_analytical(t,s0,k_off,t0=0)

    # Check if the result is a numpy array
    assert isinstance(result, np.ndarray), "The result should be a numpy array."

    # Check the length of the result
    assert len(result) == len(t), "The length of the result should match the length of t."

    # Check the first value of the result
    assert np.isclose(result[0], s0), "The first value of the result should match the initial signal."

    # Check decreasing behavior
    assert np.all(np.diff(result) <= 0), "The result should be non-increasing with increasing time."

def test_ode_one_site_mass_transport_association():

    # Parameters
    t = np.linspace(0, 100, 50)  # time points
    s1_0 = 0.0                   # initial signal
    cs_0 = 0.5                   # initial analyte surface concentration
    analyte_conc = 0.5           # high analyte concentration
    Kd = 1e-6                    # equilibrium dissociation constant
    k_off = 0.1                  # dissociation rate constant
    k_tr = 10.0                  # mass transport rate constant
    s_max = 1.0                  # maximum signal
    t0 = 0

    # Solve ODE
    signal = solve_ode_one_site_mass_transport_association(
        t, s1_0, cs_0, analyte_conc, Kd, k_off, k_tr, s_max, t0
    )

    # Check output type
    assert isinstance(signal, np.ndarray), "The output should be a numpy array."
    # Check monotonic increase
    assert np.all(np.diff(signal) >= -1e-8), "Signal should be non-decreasing with time."
    # Check initial value
    assert np.isclose(signal[0], s1_0), "Signal at t=0 should match initial value."
    # Check saturation at high concentration (should approach s_max)
    assert np.isclose(signal[-1], s_max, atol=1e-2), "Signal at high t and high analyte concentration should approach s_max."

def test_ode_one_site_mass_transport_dissociation():

    # Parameters
    t = np.linspace(0, 100, 50)  # time points
    s1_0 = 0.8                   # initial signal (start at max)
    Kd = 0.5                    # equilibrium dissociation constant
    k_off = 0.1                  # dissociation rate constant
    k_tr = 0.1                  # mass transport rate constant
    s_max = 1.0                  # maximum signal
    t0 = 0

    # Solve ODE
    signal = solve_ode_one_site_mass_transport_dissociation(
        t, s1_0, Kd, k_off, k_tr, s_max, t0
    )

    # Check output type
    assert isinstance(signal, np.ndarray), "The output should be a numpy array."
    # Check monotonic decrease
    assert np.all(np.diff(signal) <= 1e-8), "Signal should be non-increasing with time."
    # Check initial value
    assert np.isclose(signal[0], s1_0), "Signal at t=0 should match initial value."
    # Check that the signal approaches zero at long times
    assert signal[-1] < 0.05, "Signal at long times should approach zero."

def test_solve_induced_fit_association():

    # Parameters
    t = np.linspace(0, 0.5, 5)
    a_conc = 1.0
    kon = 100
    koff = 100
    kc = 10
    krev = 1
    sP1L = 0.0
    sP2L = 0.0
    smax = 1.0

    arr = solve_induced_fit_association(t, a_conc, kon, koff, kc, krev, sP1L, sP2L, smax)

    # Check output type and shape
    assert isinstance(arr, np.ndarray), "Output should be a numpy array."
    assert arr.shape == (len(t), 3), "Output should have shape (len(t), 3)."

    total_signal = arr[:, 1] + arr[:, 2]

    # Check that the signal is non-decreasing in the first column (total signal)
    assert np.all(np.diff(total_signal) >= -1e-8), "Signal should be non-decreasing with time."

    # Check that the signal starts at zero 
    assert np.isclose(total_signal[0], 0, atol=1e-6), "Initial signal should be zero."
    
    # Assuming infinite concentration of A, the last value should be close to smax
    arr = solve_induced_fit_association(t, 1e8, kon, koff, kc, krev, sP1L, sP2L, smax)
    total_signal = arr[:, 1] + arr[:, 2]

    assert np.isclose(total_signal[-1], smax, atol=1e-2), "Total signal at long times should approach smax."

def test_solve_induced_fit_dissociation():

    # Parameters
    t = np.linspace(0, 0.5, 5)
    koff = 100
    kc = 10
    krev = 1
    sP2L = 0.5
    smax = 10
    s0 = 0.7

    arr = solve_induced_fit_dissociation(t, koff, kc, krev, s0=s0, sP2L=sP2L, smax=smax)

    # Check output type and shape
    assert isinstance(arr, np.ndarray), "Output should be a numpy array."
    assert arr.shape == (len(t), 3), "Output should have shape (len(t), 3)."

    total_signal = arr[:, 1] + arr[:, 2]

    # Check that the signal is decreasing in the first column (total signal)
    assert np.all(np.diff(total_signal) <= 1e-6), "Signal should be non-increasing with time."

    # Check that the signal starts at s0 
    assert np.isclose(total_signal[0], s0, atol=1e-6), "Initial signal should match s0."
    
    # Assuming infinite time, the signal should approach zero
    arr = solve_induced_fit_dissociation([0,1,1e2,1e8], koff, kc, krev, s0=s0, sP2L=sP2L, smax=smax)

    total_signal = arr[:, 1] + arr[:, 2]

    assert np.isclose(total_signal[-1], 0, atol=1e-6), "Total signal at long times should approach zero."

def test_solve_conformational_selection_association():

    time = np.linspace(0, 1, 10)
    a_conc = 1.0
    kon = 10
    koff = 10
    kc = 10
    krev = 10
    smax = 1

    P_tot = 2

    # Compute the conformational equilibrium constant from kc and krev
    Kc = kc / krev

    # Compute the initial concentrations of P1 and P2
    P1_conc = P_tot / (1 + Kc)

    result = solve_conformational_selection_association(time, a_conc, kon, koff, kc, krev,smax=smax,sP1=P1_conc,sP2L=0)

    # Check output type and shape
    assert isinstance(result, np.ndarray), "Output should be a numpy array."
    assert result.shape == (len(time), 3), "Output should have shape (len(time), 3)."

    # Check that the signal is non-decreasing in the first column (total signal)
    total_signal = result[:, 0] 
    assert np.all(np.diff(total_signal) >= -1e-8), "Signal should be non-decreasing with time."

    # Check that at infinite concentration of A, the last value should be close to smax
    result = solve_conformational_selection_association(time, 1e8, kon, koff, kc, krev,smax=smax,sP1=P1_conc,sP2L=0)
    total_signal = result[:, 0]
    assert np.isclose(total_signal[-1], smax, atol=1e-2), "Total signal at long times should approach smax."

def test_solve_conformational_selection_dissociation():
    
    time = np.linspace(0, 1, 10)
    koff = 10
    kc = 10
    krev = 10
    smax = 10

    result = solve_conformational_selection_dissociation(time, koff, kc, krev,smax=smax,sP1=2,sP2L=5)

    # Check output type and shape
    assert isinstance(result, np.ndarray), "Output should be a numpy array."
    assert result.shape == (len(time), 3), "Output should have shape (len(time), 3)."

    total_signal = result[:, 0] 

    # Check that the signal is decreasing in the first column (total signal)
    assert np.all(np.diff(total_signal) <= 1e-6), "Signal should be non-increasing with time."

    
    # Assuming infinite time, the signal should approach zero
    result = solve_conformational_selection_dissociation([0,1e2,1e8], koff, kc, krev,smax=smax,sP1=2,sP2L=5)

    total_signal = result[:, 0]

    assert np.isclose(total_signal[-1], 0, atol=1e-6), "Total signal at long times should approach zero."

def test_solve_ode_mixture_analyte_association():

    t = np.linspace(0, 200, 100)
    Ris0 = [0,0]
    C_tot = 0.1
    Fis = np.array([0.5, 0.5])  # Example values for Fis
    Ris_max = np.array([1.0, 2])  # Example values for Ris_max
    koffs = np.array([0.1, 0.2])  # Example values for koffs
    Kds = np.array([0.5, 0.1])    # Example values for Kds

    result = solve_ode_mixture_analyte_association(t,Ris0,C_tot,Fis,Ris_max,koffs,Kds,t0=0)

    # Check output type and shape
    assert isinstance(result, np.ndarray), "Output should be a numpy array."
    assert result.shape == (len(Fis), len(t)), "Output should have shape (len(Fis), len(t))."

    # Check that the first value matches Ris0
    assert np.allclose(result[:, 0], Ris0), "The first value of the result should match Ris0."

    # Sum the responses - column axis
    total_response = np.sum(result, axis=0)

    # Check that the response increases with time
    assert np.all(np.diff(total_response) >= -1e-4), "The response should be non-decreasing with increasing time."

    # Test case with Fis[0] = 0 and infinite C_tot. Then the max value should be close to Ris_max[1]
    result = solve_ode_mixture_analyte_association(t, Ris0, 1e8, [0,1], Ris_max, koffs, Kds, t0=0)

    # Sum the responses - column axis
    total_response = np.sum(result, axis=0)

    assert np.isclose(total_response[-1], Ris_max[1], atol=1e-2), "The last values should approach the sum of Ris_max."

    # Test case with Fis[1] = 0 and infinite C_tot. Then the max value should be close to Ris_max[0]
    result = solve_ode_mixture_analyte_association(t, Ris0, 1e8, [1,0], Ris_max, koffs, Kds, t0=0)
    # Sum the responses - column axis
    total_response = np.sum(result, axis=0) 
    assert np.isclose(total_response[-1], Ris_max[0], atol=1e-2), "The last values should approach the sum of Ris_max."

def test_solve_ode_mixture_analyte_dissociation():

    Ris0 = [100, 20]
    koffs = np.array([0.1, 0.2])  # Example values for koffs
    t = np.linspace(0, 200, 100)

    # Solve the ODE for mixture analyte dissociation
    result = solve_ode_mixture_analyte_dissociation(t,Ris0,koffs,t0=0)

    # Check output type and shape
    assert isinstance(result, np.ndarray), "Output should be a numpy array."
    assert result.shape == (len(Ris0), len(t)), "Output should have shape (len(Ris0), len(t))."

    # Check that the first value matches Ris0
    assert np.allclose(result[:, 0], Ris0), "The first value of the result should match Ris0."

    # Check that the response decreases with time
    total_response = np.sum(result, axis=0)
    assert np.all(np.diff(total_response) <= 1e-4), "The response should be non-increasing with increasing time."

    # Check that the last value approaches zero - infinite time
    result = solve_ode_mixture_analyte_dissociation(np.array([0,1e2,1e8]), Ris0, koffs, t0=0)
    total_response = np.sum(result, axis=0)

    assert np.isclose(total_response[-1], 0, atol=1e-2), "The last value of the response should approach zero."


import pytest
import numpy as np

from pykingenie.utils.fitting_general import fit_many_double_exponential
from pykingenie.utils.math import double_exponential

from numpy.testing import assert_allclose

t = np.linspace(0, 10, 100)

a0 = 1
a1 = 0.1
a2 = 0.2
kobs1 = 0.5
kobs2 = 5

y = double_exponential(t, a0, a1, kobs1, a2, kobs2)

def test_fit_many_double_exponential():

    k_obs_1_fit, k_obs_2_fit, _,_,_ = fit_many_double_exponential(
        signal_lst=[y],
        time_lst=[t]
    )

    # assert that k_obs_1 is a list of NaN values - same length as signal_lst
    assert len(k_obs_1_fit) == 1, "k_obs_1 should have the same length as signal_lst."

    assert_allclose(k_obs_1_fit[0],kobs1,rtol=0.1)

    assert_allclose(k_obs_2_fit[0],kobs2,rtol=0.1)


def test_fit_many_double_exponential_no_data():

    k_obs_1, _, _, _,_ = fit_many_double_exponential(
        signal_lst=[1,2],
        time_lst=[1,2]
    )

    # assert that k_obs_1 is a list of NaN values - same length as signal_lst
    assert len(k_obs_1) == 2, "k_obs_1 should have the same length as signal_lst."

    # Verift that the first value is NaN
    assert k_obs_1[0] is np.nan, "The first value of k_obs_1 should be NaN."


import numpy as np
import pandas as pd

from .utils.fitting_general import (
    fit_many_double_exponential,
    fit_single_exponential,
    re_fit
)

from .utils.fitting_solution import (
    fit_one_site_solution,
    fit_induced_fit_solution,
    find_initial_parameters_induced_fit_solution, find_initial_parameters_conformational_selection_solution,
    fit_conformational_selection_solution
)

from .fitter import KineticsFitterGeneral

class KineticsFitterSolution(KineticsFitterGeneral):
    """
    A class used to fit solution-based kinetics data with shared thermodynamic parameters.
    
    Parameters
    ----------
    name : str
        Name of the experiment.
    assoc : list
        List containing the association signals.
    lig_conc : list
        List of ligand concentrations, one per element in assoc.
    protein_conc : list
        List of protein concentrations, one per element in assoc.
    time_assoc : list
        List of time points for the association signals.
        
    Attributes
    ----------
    name : str
        Name of the experiment.
    assoc_lst : list
        List containing the association signals.
    lig_conc : list
        List of ligand concentrations, one per element in assoc.
    prot_conc : list
        List of protein concentrations, one per element in assoc.
    time_assoc_lst : list
        List of time points for the association signals.
    signal_ss : list
        List of steady state signals.
    signal_ss_fit : list
        List of steady state fitted signals.
    signal_assoc_fit : list
        List of association kinetics fitted signals.
    fit_params_kinetics : pd.DataFrame
        DataFrame with the fitted parameters.
    fit_params_ss : pd.DataFrame
        DataFrame with the values of the fitted parameters - steady state.
    """

    def __init__(self, name, assoc, lig_conc, protein_conc, time_assoc):
        """
        Initialize the KineticsFitterSolution class.
        
        Parameters
        ----------
        name : str
            Name of the experiment.
        assoc : list
            List containing the association signals.
        lig_conc : list
            List of ligand concentrations, one per element in assoc.
        protein_conc : list
            List of protein concentrations, one per element in assoc.
        time_assoc : list
            List of time points for the association signals.
        """

        super().__init__()

        self.name  = name
        self.assoc_lst = assoc

        self.lig_conc     = lig_conc
        self.prot_conc    = protein_conc

        self.time_assoc_lst = time_assoc

    def get_steady_state(self):
        """
        Get the steady state signals from the association signals.
        
        The steady state signal is calculated as the median of the last 5 values
        of each association signal. The signals are grouped by protein concentration.
        
        Returns
        -------
        None
            Updates the following instance attributes:
            - unq_prot_conc: Unique protein concentrations
            - signal_ss_per_protein: Steady state signals grouped by protein concentration
            - lig_conc_per_protein: Ligand concentrations grouped by protein concentration
        """

        signal_ss_per_protein   = []
        ligand_conc_per_protein = []

        unq_prot_conc = pd.unique(np.array(self.prot_conc)) # Follow the order of appearance in the list

        for prot in unq_prot_conc:

            # Get the indices of the association signals for the current protein concentration
            indices = [i for i, p in enumerate(self.prot_conc) if p == prot]

            # Get the corresponding association signals
            assoc_signals = [self.assoc_lst[i] for i in indices]

            # Get the corresponding ligand concentrations
            lig_conc = [self.lig_conc[i] for i in indices]

            # Calculate the steady state signal as the median of the last 5 values of each signal
            signal_ss = [np.median(y[-5:]) if len(y) >= 5 else np.nan for y in assoc_signals]

            signal_ss_per_protein.append(np.array(signal_ss))
            ligand_conc_per_protein.append(np.array(lig_conc))

        self.unq_prot_conc         = unq_prot_conc
        self.signal_ss_per_protein = signal_ss_per_protein
        self.lig_conc_per_protein  = ligand_conc_per_protein

        return None

    def fit_single_exponentials(self):

        """
        Fit single exponentials to the association signals in the solution kinetics experiment.
        
        This method fits each association curve with a single exponential function
        and extracts the observed rate constants (k_obs). The results are then grouped
        by protein concentration.
        
        Returns
        -------
        None
            Updates the following instance attributes:
            - k_obs: List of observed rate constants for each association curve
            - signal_assoc_fit: List of fitted association signals
            - fit_params_kinetics: DataFrame with the fitted parameters
        """
        self.clear_fittings()

        k_obs  = [np.nan for _ in range(len(self.assoc_lst))]
        y_pred = [None for _ in range(len(self.assoc_lst))]

        i = 0
        for y,t in zip(self.assoc_lst,self.time_assoc_lst):

            fit_params, cov, fit_y = fit_single_exponential(y,t)

            k_obs[i] = fit_params[2]
            y_pred[i] = fit_y

            i += 1

        self.k_obs            = k_obs
        self.signal_assoc_fit = y_pred

        self.group_k_obs_by_protein_concentration()

        # Create a DataFrame with the fitted parameters and assign it to fit_params_kinetics
        self.fit_params_kinetics = pd.DataFrame({
            'Protein [µM]': self.prot_conc,
            'Ligand [µM]':  self.lig_conc,
            'k_obs [1/s]':  self.k_obs
        })

        return None

    def group_k_obs_by_protein_concentration(self):
        """
        Group the observed rate constants by protein concentration.
        
        This is useful for plotting the rate constants against the protein concentration
        and for subsequent analysis of the concentration dependence of the kinetics.
        
        Returns
        -------
        None
            Updates the k_obs_per_prot attribute with a dictionary mapping
            protein concentrations to arrays of observed rate constants.
        """
        k_obs_per_prot = {}

        for prot, k in zip(self.prot_conc, self.k_obs):
            if prot not in k_obs_per_prot:
                k_obs_per_prot[prot] = []
            k_obs_per_prot[prot].append(k)

        # Convert to numpy arrays for easier handling later
        for prot in k_obs_per_prot:
            k_obs_per_prot[prot] = np.array(k_obs_per_prot[prot])

        self.k_obs_per_prot = k_obs_per_prot

        return None

    def fit_double_exponentials(self, min_log_k=-4, max_log_k=4, log_k_points=22):

        """
        Fit double exponentials to the association signals in the solution kinetics experiment.
        
        This method fits each association curve with a double exponential function
        and extracts two observed rate constants (k_obs_1 and k_obs_2). The results 
        are then grouped by protein concentration.
        
        Parameters
        ----------
        min_log_k : float, optional
            Minimum value of log10(k) to search, default is -4.
        max_log_k : float, optional
            Maximum value of log10(k) to search, default is 4.
        log_k_points : int, optional
            Number of points to sample in the log(k) space, default is 22.
            
        Returns
        -------
        None
            Updates the following instance attributes:
            - k_obs_1: List of first observed rate constants for each association curve
            - k_obs_2: List of second observed rate constants for each association curve
            - signal_assoc_fit: List of fitted association signals
            - fit_params_kinetics: DataFrame with the fitted parameters
        """
        self.clear_fittings()

        double_exp_results = fit_many_double_exponential(self.assoc_lst,self.time_assoc_lst,min_log_k, max_log_k, log_k_points)

        k_obs_1, k_obs_2, y_pred, k_obs_1_err, k_obs_2_err = double_exp_results

        self.k_obs_1 = k_obs_1
        self.k_obs_2 = k_obs_2
        self.signal_assoc_fit = y_pred

        self.k_obs_1_err = k_obs_1_err
        self.k_obs_2_err = k_obs_2_err

        self.group_double_exponential_k_obs_by_protein_concentration()

        # Create a DataFrame with the fitted parameters and assign it to fit_params_kinetics
        self.fit_params_kinetics = pd.DataFrame({
            'Protein [µM]': self.prot_conc,
            'Ligand [µM]':  self.lig_conc,
            'k_obs_1 [1/s]': self.k_obs_1,
            'k_obs_1 relative error (%)': self.k_obs_1_err,
            'k_obs_2 [1/s]': self.k_obs_2,
            'k_obs_2 relative error (%)': self.k_obs_2_err
        })

        return None

    def group_double_exponential_k_obs_by_protein_concentration(self):

        """
        Group the observed rate constants by protein concentration for double exponential fits.
        
        This is useful for plotting the rate constants against the protein concentration
        and for subsequent analysis of the concentration dependence of the kinetics
        when using double exponential fits.
        
        Returns
        -------
        None
            Updates the following instance attributes:
            - k_obs_1_per_prot: Dictionary mapping protein concentrations to arrays of first observed rate constants
            - k_obs_2_per_prot: Dictionary mapping protein concentrations to arrays of second observed rate constants
        """
        k_obs_1_per_prot = {}
        k_obs_2_per_prot = {}

        for prot, k1, k2 in zip(self.prot_conc, self.k_obs_1, self.k_obs_2):
            if prot not in k_obs_1_per_prot:
                k_obs_1_per_prot[prot] = []
                k_obs_2_per_prot[prot] = []
            k_obs_1_per_prot[prot].append(k1)
            k_obs_2_per_prot[prot].append(k2)

        # Convert to numpy arrays for easier handling later
        for prot in k_obs_1_per_prot:
            k_obs_1_per_prot[prot] = np.array(k_obs_1_per_prot[prot])
            k_obs_2_per_prot[prot] = np.array(k_obs_2_per_prot[prot])

        self.k_obs_1_per_prot = k_obs_1_per_prot
        self.k_obs_2_per_prot = k_obs_2_per_prot

        return None

    def fit_one_binding_site(
            self,fit_signal_E=False,
            fit_signal_S=False,
            fit_signal_ES=True,
            fixed_t0=True):
        """
        Fit the association signals assuming one binding site.
        
        This is a simplified model that assumes a single binding site for the ligand 
        on the protein. The model fits global parameters to all curves simultaneously.

        Parameters
        ----------
        fit_signal_E : bool, optional
            If True, fit the signal of the free protein E, default is False.
        fit_signal_S : bool, optional
            If True, fit the signal of the free ligand S, default is False.
        fit_signal_ES : bool, optional
            If True, fit the signal of the complex ES, default is True.
        fixed_t0 : bool, optional
            If True, fix the t0 parameter to 0, default is True.

        Returns
        -------
        None
            Updates the following instance attributes:
            - signal_assoc_fit: List of fitted association signals
            - fit_params_kinetics: DataFrame with the fitted parameters
        """

        # Initial parameters are as follows:
        # 1. Signal amplitude of the free protein
        # 2. Signal amplitude of the free ligand
        # 3. Signal amplitude of the complex
        # 4. Kd of the complex
        # 5. koff of the complex
        # 6. t0 of the first trace
        # 7. t0 of the second trace
        # n. t0 of the last trace

        self.clear_fittings()

        max_signal = np.max(self.assoc_lst)
        max_prot   = np.max(self.prot_conc)
        max_lig    = np.max(self.lig_conc)

        max_signal_P = max_signal / max_prot
        max_signal_L = max_signal / max_lig

        min_Kd =  np.min(self.lig_conc)/1e2
        max_Kd = np.max(self.lig_conc)*1e2

        initial_parameters = [max_signal_P,max_signal_L,max_signal_P,np.mean(self.lig_conc),1]
        low_bounds  = [0,0,0,min_Kd,1e-4]
        high_bounds = [np.inf,np.inf,np.inf,max_Kd,1e5]

        # Remove the third parameter if we do not fit the signal of the complex
        if not fit_signal_ES:
            initial_parameters.pop(2)
            low_bounds.pop(2)
            high_bounds.pop(2)

        # Remove the second parameter if we do not fit the signal of the free ligand
        if not fit_signal_S:
            initial_parameters.pop(1)
            low_bounds.pop(1)
            high_bounds.pop(1)

        # Remove the first parameter if we do not fit the signal of the free protein
        if not fit_signal_E:
            initial_parameters.pop(0)
            low_bounds.pop(0)
            high_bounds.pop(0)

        # If t0 is fitted, we need to include it in the initial parameters
        if not fixed_t0:

            for _ in range(len(self.assoc_lst)):

                initial_parameters.append(0)
                low_bounds.append(-0.1)
                high_bounds.append(0.1)

        fit_params, cov, fitted_values, parameter_names = fit_one_site_solution(
            signal_lst=self.assoc_lst,
            time_lst=self.time_assoc_lst,
            ligand_conc_lst=self.lig_conc,
            protein_conc_lst=self.prot_conc,
            initial_parameters=initial_parameters,
            low_bounds=low_bounds,
            high_bounds=high_bounds,
            fit_signal_E=fit_signal_E,
            fit_signal_S=fit_signal_S,
            fit_signal_ES=fit_signal_ES,
            fixed_t0=fixed_t0,
            fixed_Kd=False,
            Kd_value=None,
            fixed_koff=False,
            koff_value=None
        )

        kwargs = {
            'signal_lst': self.assoc_lst,
            'time_lst': self.time_assoc_lst,
            'ligand_conc_lst': self.lig_conc,
            'protein_conc_lst': self.prot_conc,
            'fit_signal_E': fit_signal_E,
            'fit_signal_S': fit_signal_S,
            'fit_signal_ES': fit_signal_ES,
            'fixed_t0': fixed_t0
        }

        # Re-fit the parameters if they are close to the constraints
        fit_params, cov, fit_vals, low_bounds, high_bounds = re_fit(
            fit=fit_params,
            cov=cov,
            fit_vals=fitted_values,
            fit_fx=fit_one_site_solution,
            low_bounds=low_bounds,
            high_bounds=high_bounds,
            times = 2,
            **kwargs
        )

        self.params = fit_params
        self.low_bounds  = low_bounds
        self.high_bounds = high_bounds
        self.create_fitting_bounds_table()

        self.signal_assoc_fit = fit_vals

        # Create a DataFrame with the fitted parameters and assign it to fit_params_kinetics

        self.fit_params_kinetics = pd.DataFrame({
            'Protein [µM]': self.prot_conc,
            'Ligand [µM]':  self.lig_conc,
        })

        for i, param in enumerate(parameter_names):

            if "t0" not in param:

                self.fit_params_kinetics[param] = fit_params[i]

        # If we fit t0, we need to restructure the fit_params_kinetics DataFrame
        if not fixed_t0:
            n_t0 = len(self.assoc_lst)
            t0_params = fit_params[-n_t0:]

            # Include a column named t0
            self.fit_params_kinetics['t0'] = t0_params

        return None

    def find_initial_params_if(self,
                               fit_signal_E=False,
                               fit_signal_S=False,
                               fit_signal_ES=True,
                               ESint_equals_ES=True,
                               fixed_t0=True,
                               fit_scale_factor=False,
                               scale_factor_tolerance=0.2
                               ):
        """
        Find optimal initial parameters for the induced fit model.
        
        Heuristically finds the best initial parameters for the fit by exploring
        fixed values of kc and krev and fitting kon and koff (and the signal of the complex).
        
        Parameters
        ----------
        fit_signal_E : bool, optional
            If True, fit the signal of the free protein E, default is False.
        fit_signal_S : bool, optional
            If True, fit the signal of the free ligand S, default is False.
        fit_signal_ES : bool, optional
            If True, fit the signal of the complex ES, default is True.
        ESint_equals_ES : bool, optional
            If True, assume that the signal of the intermediate ESint is equal 
            to the signal of the complex ES, default is True.
        fixed_t0 : bool, optional
            If True, fix the t0 parameter to 0, default is True.
        fit_scale_factor : bool, optional
            If True, fit a scale factor for the signals, default is False.
        scale_factor_tolerance : float, optional
            Tolerance for the scale factor fit, default is 0.2.
            
        Returns
        -------
        None
            Updates the params_guess attribute with the best initial parameters.
        """

        self.clear_fittings()

        # Heuristically find the best initial parameters for the fit
        # We explore fixed values of kc and krev and fit kon and koff (and the signal of the complex)
        params_guess = find_initial_parameters_induced_fit_solution(
            signal_lst=self.assoc_lst,
            time_lst=self.time_assoc_lst,
            ligand_conc_lst=self.lig_conc,
            protein_conc_lst=self.prot_conc,
            fit_signal_E=fit_signal_E,
            fit_signal_S=fit_signal_S,
            fit_signal_ES=fit_signal_ES,
            ESint_equals_ES=ESint_equals_ES,
            fixed_t0=fixed_t0,
            fit_scale_factor=fit_scale_factor,
            scale_factor_tolerance=scale_factor_tolerance)

        self.params_guess = params_guess

        return None

    def fit_induced_fit(self,
                        fit_signal_E=False,
                        fit_signal_S=False,
                        fit_signal_ES=True,
                        ESint_equals_ES=True,
                        fixed_t0=True,
                        fit_scale_factor=False,
                        scale_factor_tolerance=0.2
                        ):
        """
        Fit the association signals assuming an induced-fit mechanism.

        A + B <-> AB_intermediate <-> AB_trapped

        This model accounts for conformational changes in the protein upon ligand binding.

        Parameters
        ----------
        fit_signal_E : bool, optional
            If True, fit the signal of the free protein E, default is False.
        fit_signal_S : bool, optional
            If True, fit the signal of the free ligand S, default is False.
        fit_signal_ES : bool, optional
            If True, fit the signal of the complex ES, default is True.
        ESint_equals_ES : bool, optional
            If True, assume that the signal of the intermediate ESint is equal to the signal of the complex ES, default is True.
        fixed_t0 : bool, optional
            If True, fix the t0 parameter to 0, default is True.
        fit_scale_factor : bool, optional
            If True, fit a scale factor for the signals, default is False.
        scale_factor_tolerance : float, optional
            Tolerance for the scale factor fit, default is 0.2.
            
        Returns
        -------
        None
            Updates the following instance attributes:
            - signal_assoc_fit: List of fitted association signals
            - fit_params_kinetics: DataFrame with the fitted parameters
        """

        # fit using as initial parameters the best found parameters
        initial_parameters = np.array(self.params_guess )
        low_bounds  = [x / 1e3 if x > 0 else x*1e3 for x in initial_parameters ]
        high_bounds = [x * 1e3 if x > 0 else x*1e-3 for x in initial_parameters ]

        # Set the bounds for the t0 parameter - between -0.1 and 0.1
        if not fixed_t0:
            n_params = len(initial_parameters) - len(self.assoc_lst) * fit_scale_factor
            for i in range(len(self.assoc_lst)):
                low_bounds[n_params-i-1] = -0.1
                high_bounds[n_params-i-1] = 0.1

        # Set the bounds for the scale factor parameter - between 1-scale_factor_tolerance and 1+scale_factor_tolerance
        if fit_scale_factor:
            n_params = len(initial_parameters)
            for i in range(len(self.assoc_lst)):
                low_bounds[n_params - i - 1] = 1 - scale_factor_tolerance
                high_bounds[n_params - i - 1] = 1 + scale_factor_tolerance

        global_fit_params, cov, fitted_values, parameter_names = fit_induced_fit_solution(
            signal_lst=self.assoc_lst,
            time_lst=self.time_assoc_lst,
            ligand_conc_lst=self.lig_conc,
            protein_conc_lst=self.prot_conc,
            initial_parameters= initial_parameters,
            low_bounds=low_bounds,
            high_bounds=high_bounds,
            fit_signal_E=fit_signal_E,
            fit_signal_S=fit_signal_S,
            fit_signal_ES=fit_signal_ES,
            ESint_equals_ES=ESint_equals_ES,
            fixed_t0=fixed_t0,
            fit_scale_factor=fit_scale_factor
        )

        kwargs = {
            'signal_lst': self.assoc_lst,
            'time_lst': self.time_assoc_lst,
            'ligand_conc_lst': self.lig_conc,
            'protein_conc_lst': self.prot_conc,
            'fit_signal_E': fit_signal_E,
            'fit_signal_S': fit_signal_S,
            'fit_signal_ES': fit_signal_ES,
            'ESint_equals_ES': ESint_equals_ES,
            'fixed_t0': fixed_t0,
            'fit_scale_factor': fit_scale_factor
        }

        # Re-fit the parameters if they are close to the constraints
        global_fit_params, cov, fitted_values, low_bounds, high_bounds = re_fit(
            fit=global_fit_params,
            cov=cov,
            fit_vals=fitted_values,
            fit_fx=fit_induced_fit_solution,
            low_bounds=low_bounds,
            high_bounds=high_bounds,
            times=2,
            **kwargs
        )

        self.params  = global_fit_params
        self.low_bounds  = low_bounds
        self.high_bounds = high_bounds

        self.create_fitting_bounds_table()

        self.signal_assoc_fit = fitted_values

        # Create a DataFrame with the fitted parameters and assign it to fit_params_kinetics
        self.fit_params_kinetics = pd.DataFrame({
            'Protein [µM]': self.prot_conc,
            'Ligand [µM]':  self.lig_conc,
        })

        for i, param in enumerate(parameter_names):
            if "t0" not in param:
                self.fit_params_kinetics[param] = global_fit_params[i]

        # Include the t0 column in the dataframe if we fitted t0
        if not fixed_t0:
            n_t0 = len(self.assoc_lst)
            t0_params = global_fit_params[-n_t0:]
            self.fit_params_kinetics['t0'] = t0_params

        return None

    def find_initial_params_cs(
            self,
            fit_signal_E=False,
            E1_equals_E2=True,
            fit_signal_S=False,
            fit_signal_E2S=True,
            fixed_t0=True,
            fit_scale_factor=False,
            scale_factor_tolerance=0.2):
        """
        Find optimal initial parameters for the conformational selection model.

        Heuristically finds the best initial parameters for the fit by exploring
        fixed values of kc and krev and fitting kon and koff (and the signal of the complex).

        Parameters
        ----------
        fit_signal_E : bool, optional
            If True, fit the signal of the free protein E, default is False.
        E1_equals_E2 : bool, optional
            If True, assume that the signal of the two conformations of the free protein E1
            and E2 are equal, default is True.
        fit_signal_S : bool, optional
            If True, fit the signal of the free ligand S, default is False.
        fit_signal_E2S : bool, optional
            If True, fit the signal of the complex E2S, default is True.
        fixed_t0 : bool, optional
            If True, fix the t0 parameter to 0, default is True.
        fit_scale_factor : bool, optional
            If True, fit a scale factor for the signals, default is False.
        scale_factor_tolerance : float, optional
            Tolerance for the scale factor fit, default is 0.2.

        Returns
        -------
        None
            Updates the params_guess attribute with the best initial parameters.
        """

        self.clear_fittings()

        # Heuristically find the best initial parameters for the fit
        # We explore fixed values of kc and krev and fit kon and koff (and the signal of the complex)
        params_guess = find_initial_parameters_conformational_selection_solution(
            signal_lst=self.assoc_lst,
            time_lst=self.time_assoc_lst,
            ligand_conc_lst=self.lig_conc,
            protein_conc_lst=self.prot_conc,
            fit_signal_E=fit_signal_E,
            E1_equals_E2=E1_equals_E2,
            fit_signal_S=fit_signal_S,
            fit_signal_E2S=fit_signal_E2S,
            fixed_t0=fixed_t0,
            fit_scale_factor=fit_scale_factor,
            scale_factor_tolerance=scale_factor_tolerance
            )
        
        self.params_guess = params_guess

        return None

    def fit_conformational_selection(
            self,
            fit_signal_E=False,
            E1_equals_E2=True,
            fit_signal_S=False,
            fit_signal_E2S=True,
            fixed_t0=True,
            fit_scale_factor=False,
            scale_factor_tolerance=0.2
            ):
        """
        Fit the association signals assuming an conformational selection mechanism.

        A_active <-> A_inactive

        A_active + B <-> AB

        This model accounts for conformational changes in the protein before ligand binding.

        Parameters
        ----------
        fit_signal_E : bool, optional
            If True, fit the signal of the free protein E, default is False.
        E1_equals_E2 : bool, optional
            If True, assume that the signal of the two conformations of the free protein E1
            and E2 are equal, default is True.
        fit_signal_S : bool, optional
            If True, fit the signal of the free ligand S, default is False.
        fit_signal_E2S : bool, optional
            If True, fit the signal of the complex E2S, default is True.
        fixed_t0 : bool, optional
            If True, fix the t0 parameter to 0, default is True.

        Returns
        -------
        None
            Updates the following instance attributes:
            - signal_assoc_fit: List of fitted association signals
            - fit_params_kinetics: DataFrame with the fitted parameters
        """

        # fit using as initial parameters the best found parameters
        initial_parameters = np.array(self.params_guess)
        low_bounds = [x / 1e3 if x > 0 else x * 1e3 for x in initial_parameters]
        high_bounds = [x * 1e3 if x > 0 else x * 1e-3 for x in initial_parameters]

        # Set the bounds for the t0 parameter - between -0.1 and 0.1
        if not fixed_t0:
            n_params = len(initial_parameters) - len(self.assoc_lst) * fit_scale_factor
            for i in range(len(self.assoc_lst)):
                low_bounds[n_params - i - 1] = -0.1
                high_bounds[n_params - i - 1] = 0.1

        # Set the bounds for the scale factor parameter - between 1-scale_factor_tolerance and 1+scale_factor_tolerance
        if fit_scale_factor:
            n_params = len(initial_parameters)
            for i in range(len(self.assoc_lst)):
                low_bounds[n_params - i - 1] = 1 - scale_factor_tolerance
                high_bounds[n_params - i - 1] = 1 + scale_factor_tolerance


        global_fit_params, cov, fitted_values, parameter_names = fit_conformational_selection_solution(
            signal_lst=self.assoc_lst,
            time_lst=self.time_assoc_lst,
            ligand_conc_lst=self.lig_conc,
            protein_conc_lst=self.prot_conc,
            initial_parameters=initial_parameters,
            low_bounds=low_bounds,
            high_bounds=high_bounds,
            fit_signal_E=fit_signal_E,
            E1_equals_E2=E1_equals_E2,
            fit_signal_S=fit_signal_S,
            fit_signal_E2S=fit_signal_E2S,
            fixed_t0=fixed_t0,
            fit_scale_factor=fit_scale_factor
        )

        kwargs = {
            'signal_lst': self.assoc_lst,
            'time_lst': self.time_assoc_lst,
            'ligand_conc_lst': self.lig_conc,
            'protein_conc_lst': self.prot_conc,
            'fit_signal_E': fit_signal_E,
            'E1_equals_E2' : E1_equals_E2,
            'fit_signal_S': fit_signal_S,
            'fit_signal_E2S': fit_signal_E2S,
            'fixed_t0': fixed_t0,
            'fit_scale_factor': fit_scale_factor
        }

        # Re-fit the parameters if they are close to the constraints
        global_fit_params, cov, fitted_values, low_bounds, high_bounds = re_fit(
            fit=global_fit_params,
            cov=cov,
            fit_vals=fitted_values,
            fit_fx=fit_conformational_selection_solution,
            low_bounds=low_bounds,
            high_bounds=high_bounds,
            times=2,
            **kwargs
        )

        self.params = global_fit_params
        self.low_bounds = low_bounds
        self.high_bounds = high_bounds

        self.create_fitting_bounds_table()

        self.signal_assoc_fit = fitted_values

        # Create a DataFrame with the fitted parameters and assign it to fit_params_kinetics
        self.fit_params_kinetics = pd.DataFrame({
            'Protein [µM]': self.prot_conc,
            'Ligand [µM]': self.lig_conc,
        })

        for i, param in enumerate(parameter_names):
            if "t0" not in param:
                self.fit_params_kinetics[param] = global_fit_params[i]

        # Include the t0 column in the dataframe if we fitted t0
        if not fixed_t0:
            n_t0 = len(self.assoc_lst)
            t0_params = global_fit_params[-n_t0:]
            self.fit_params_kinetics['t0'] = t0_params

        return None

    def create_export_df(self,type='raw'):

        """
        Create


        Parameters
        ----------
        type

        Returns
        -------
    name : str
        Name of the experiment.
    assoc_lst : list
        List containing the association signals.
    lig_conc : list
        List of ligand concentrations, one per element in assoc.
    prot_conc : list
        List of protein concentrations, one per element in assoc.
    time_assoc_lst : list
        List of time points for the association signals.
        """

        df_list = []

        if type == 'raw':
            assoc_lst = self.assoc_lst
        else:
            assoc_lst = self.signal_assoc_fit

        for signal, time, lig, prot in zip(assoc_lst, self.time_assoc_lst, self.lig_conc, self.prot_conc):

            df = pd.DataFrame({
                'Time': time,
                'Signal': signal,
                'Ligand_concentration_micromolar': lig,
                'Protein_concentration_micromolar': prot
            })

            df_list.append(df)

        export_df = pd.concat(df_list, ignore_index=True)

        return export_df


import numpy as np
import pytest
import os

from pykingenie.kingenie_surface import KinGenieCsv

csv_file_single_cycle = "./test_files/single_cycle_kingenie.csv"
csv_file_multi_cycle  = "./test_files/multi_cycle_kingenie.csv"

kingenie_csv = KinGenieCsv('test_kingenie_csv')

def test_instance_creation():
    """
    Test the creation of a KinGenieCsv instance.
    """
    assert isinstance(kingenie_csv, KinGenieCsv), "The instance should be of type KinGenieCsv."
    assert kingenie_csv.name == "test_kingenie_csv"

def test_read_csv():

    kingenie_csv.read_csv(csv_file_single_cycle)

    assert len(kingenie_csv.xs) > 0, "The xs list should not be empty after reading a single cycle CSV file."

    kingenie_csv.read_csv(csv_file_multi_cycle)

    assert len(kingenie_csv.xs) > 0, "The xs list should not be empty after reading a multi-cycle CSV file."

def test_get_step_xy():
    """
    Test the get_step_xy method.
    """

    kingenie_csv.read_csv(csv_file_single_cycle)

    x = kingenie_csv.get_step_xy(kingenie_csv.sensor_names[0],
                                 location_loading=1, #  loading location for this simulated file
                                 location_sample=2,  # sample location for this simulated file
                                 step_type='ASSOC',
                                 replicate=1,
                                 type='x')
    
    # Check if x is a numpy array
    assert isinstance(x, np.ndarray), "The x should be a numpy array."

    y = kingenie_csv.get_step_xy(kingenie_csv.sensor_names[0],
                                    location_loading=1, #  loading location for this simulated file 
                                    location_sample=2,  # sample location for this simulated file
                                    step_type='ASSOC',
                                    replicate=1,
                                    type='y')
    
    # Check if y is a numpy array
    assert isinstance(y, np.ndarray), "The y should be a numpy array."

    x = kingenie_csv.get_step_xy(kingenie_csv.sensor_names[0],
                                    location_loading=1, #  loading location for this simulated file
                                    location_sample=3,  # sample location for this simulated file
                                    step_type='DISASSOC',   
                                    replicate=1,
                                    type='x')

    # Check if x is a numpy array
    assert isinstance(x, np.ndarray), "The x should be a numpy array."

    y = kingenie_csv.get_step_xy(kingenie_csv.sensor_names[0],
                                    location_loading=1, #  loading location for this simulated file
                                    location_sample=3,  # sample location for this simulated file
                                    step_type='DISASSOC',
                                    replicate=1,
                                    type='y')   
    # Check if y is a numpy array
    assert isinstance(y, np.ndarray), "The y should be a numpy array."



import pandas as pd
import numpy  as np

from .surface_exp import SurfaceBasedExperiment

class KinGenieCsv(SurfaceBasedExperiment):
    """
    A class used to represent a KinGenie surface-based simulation.
    
    This class handles data from KinGenie simulation CSV files that can be exported 
    from the Simulation panel.

    Parameters
    ----------
    name : str, optional
        Name of the experiment. Default is 'kingenie_csv'.

    Attributes
    ----------
    name : str
        Name of the experiment.
    fn : str
        File name.
    xs : list of list of numpy.ndarray
        List of time values for each sensor and phase. Structure is:
        [sensor_1[assoc_phase, disso_phase], sensor_2[assoc_phase, disso_phase], ...].
    ys : list of list of numpy.ndarray
        List of signal values for each sensor and phase. Structure is:
        [sensor_1[assoc_phase, disso_phase], sensor_2[assoc_phase, disso_phase], ...].
    no_sensors : int
        Number of sensors.
    sensor_names : list of str
        List of sensor names, one per sensor.
    sensor_names_unique : list of str
        List of unique sensor names.
    ligand_conc_df : pandas.DataFrame
        DataFrame with the ligand concentration information.
    df_steps : pandas.DataFrame
        DataFrame with step information (association, dissociation).
    traces_loaded : bool
        Whether traces have been loaded.
        
    Notes
    -----
    The CSV file should contain columns for Time, Signal, Smax, and 
    Analyte_concentration_micromolar_constant. For single cycle kinetics,
    it should also include a Cycle column.
    """

    def __init__(self, name='kingenie_csv'):

        super().__init__(name, 'kingenie_csv')

    def read_csv(self, file):
        """
        Read the KinGenie CSV file containing surface-based simulation data.
        
        Parameters
        ----------
        file : str
            Path to the CSV file.
            
        Returns
        -------
        None
            The method populates class attributes with data from the CSV file.
            
        Notes
        -----
        The CSV file should have the following columns:
        - Time: time points of the simulation
        - Signal: signal values at each time point
        - Smax: maximum signal value
        - Analyte_concentration_micromolar_constant: analyte concentration
        - Cycle: (optional) cycle number for single cycle kinetics
        
        This method creates the following attributes:
        - self.xs: list of time values for each sensor and phase
        - self.ys: list of signal values for each sensor and phase
        - self.no_sensors: number of sensors
        - self.sensor_names: list of sensor names
        - self.ligand_conc_df: DataFrame with concentration information
        - self.df_steps: DataFrame with step information
        
        The method handles both multi-cycle and single-cycle kinetics data.
        """

        df = pd.read_csv(file)

        # Read only the first Smax value
        smax_unq = df['Smax'].unique()
        df       = df[df['Smax'] == smax_unq[0]]

        # Detect if we have the column cycle
        is_single_cycle = 'Cycle' in df.columns and len(df['Cycle'].unique()) > 1

        # Generate one 'fake' sensor per ligand concentration, if it's not a single cycle
        ligand_conc = df['Analyte_concentration_micromolar_constant'].unique()

        # Remove zero concentration
        ligand_conc = ligand_conc[ligand_conc != 0]

        if is_single_cycle:
            self.no_sensors = 1
        else:
            self.no_sensors = len(ligand_conc)

        # Use fake names
        self.sensor_names = ['sim. sensor ' + str(i+1) for i in range(self.no_sensors)]

        # Initiate self.xs and self.ys, one empty list per sensor
        self.xs = [[] for _ in range(self.no_sensors)]
        self.ys = [[] for _ in range(self.no_sensors)]

        steps_start = []

        if not is_single_cycle:

            # Now populate, for each sensor self.xs and self.ys
            for i in range(self.no_sensors):

                # Find the start index of the association phase
                start_idx = df[df['Analyte_concentration_micromolar_constant'] == ligand_conc[i]].index[0]

                if i == self.no_sensors - 1:
                    end_idx = df.shape[0]
                else:
                    end_idx   = df[df['Analyte_concentration_micromolar_constant'] == ligand_conc[i+1]].index[0]

                df_temp = df.iloc[start_idx:end_idx]

                # Extract the association phase and dissociation phase
                df_temp_asso  = df_temp[df_temp['Analyte_concentration_micromolar_constant'] > 0]
                df_temp_disso = df_temp[df_temp['Analyte_concentration_micromolar_constant'] == 0]

                # Extract the time values of the association/dissociation phase
                time_asso = df_temp_asso['Time'].to_numpy()
                time_diss = df_temp_disso['Time'].to_numpy()

                if i == 1:

                    steps_start.append(0)
                    steps_start.append(time_asso[-1])

                # Populate self.xs
                self.xs[i].append(time_asso)
                self.xs[i].append(time_diss)

                # Extract the signal values
                signal_asso = df_temp_asso['Signal'].to_numpy()
                signal_diss = df_temp_disso['Signal'].to_numpy()

                # Populate self.ys
                self.ys[i].append(signal_asso)
                self.ys[i].append(signal_diss)

            # Now generate a fake ligand concentration df
            df_sensor = pd.DataFrame({'Sensor':self.sensor_names,
                                      'Analyte_location':'1',
                                      'Concentration_micromolar':ligand_conc,
                                      'SampleID':'simulation',
                                      'Replicate':1,
                                      'Loading_location': '2',
                                      'Experiment':self.name})

            self.ligand_conc_df = df_sensor

            # Generate a fake steps df
            steps_names = ['Association','Dissociation']
            steps_types = ['ASSOC','DISASSOC']
            steps_loc   = ['1','3']
            steps_load  = ['2', 'NA']

            self.df_steps = pd.DataFrame({'#Step':np.arange(len(steps_names))+1,
                                          'Name':steps_names,
                                          'Type':steps_types,
                                          'Start':steps_start,
                                          'Column_location':steps_loc,
                                          'Loading_location':steps_load})

        else:

            cycles = df['Cycle'].unique()
            cycles.sort()

            n_cycles = len(cycles)

            analyte_location = []
            steps_loc        = []

            for cycle in cycles:

                analyte_location.append(str(cycle+1))
                steps_loc.append(str(cycle+1))
                steps_loc.append(str(999))

                df_temp = df[df['Cycle'] == cycle]

                # Extract the association phase and dissociation phase
                df_temp_asso = df_temp[df_temp['Analyte_concentration_micromolar_constant'] > 0]
                df_temp_disso = df_temp[df_temp['Analyte_concentration_micromolar_constant'] == 0]

                # Extract the time values of the association/dissociation phase
                time_asso = df_temp_asso['Time'].to_numpy()
                time_diss = df_temp_disso['Time'].to_numpy()

                steps_start.append(time_asso[0])
                steps_start.append(time_asso[-1])

                # Populate self.xs
                self.xs[0].append(time_asso)
                self.xs[0].append(time_diss)

                # Extract the signal values
                signal_asso = df_temp_asso['Signal'].to_numpy()
                signal_diss = df_temp_disso['Signal'].to_numpy()

                # Populate self.ys
                self.ys[0].append(signal_asso)
                self.ys[0].append(signal_diss)

            # Now generate a fake ligand concentration df
            df_sensor = pd.DataFrame({'Sensor': self.sensor_names*n_cycles,
                                      'Analyte_location': analyte_location,
                                      'Concentration_micromolar': ligand_conc,
                                      'SampleID': 'simulation',
                                      'Replicate': 1,
                                      'Loading_location': '1',
                                      'Experiment': self.name})

            self.ligand_conc_df = df_sensor

            # Generate a fake steps df
            steps_names = ['Association', 'Dissociation'] * n_cycles
            steps_types = ['ASSOC', 'DISASSOC']           * n_cycles
            steps_load  = ['1', 'NA']                     * n_cycles

            self.df_steps = pd.DataFrame({'#Step': np.arange(len(steps_names)) + 1,
                                          'Name': steps_names,
                                          'Type': steps_types,
                                          'Start': steps_start,
                                          'Column_location': steps_loc,
                                          'Loading_location': steps_load})

        self.create_unique_sensor_names()

        self.traces_loaded = True

        return None



import pytest

from pykingenie.main import KineticsAnalyzer
from pykingenie.kingenie_solution import KinGenieCsvSolution

pyKinetics = KineticsAnalyzer()

file1 = './test_files/solution_data_1.csv'
file2 = './test_files/solution_data_2.csv'

def test_merge_solution_experiments():

    k1 = KinGenieCsvSolution('test1')
    k2 = KinGenieCsvSolution('test2')

    k1.read_csv(file1)
    k2.read_csv(file2)

    pk = KineticsAnalyzer()

    pk.add_experiment(k1, 'test1')
    pk.add_experiment(k2, 'test2')

    pk.collapse_solution_experiments()

    assert len(pk.experiments) == 1




import numpy as np
import pandas as pd
import pytest
import os

from pykingenie.utils.processing import (
    guess_experiment_name,
    guess_experiment_type,
    detect_time_list_continuos,
    get_palette,
    subset_data,
    sample_type_to_letter,
    combine_sequences,
    get_colors_from_numeric_values,
    if_string_to_list,
    find_loading_column,
    find_index_of_previous_step_type
    )

def test_guess_experiment_name():
    # Create a temporary file with known content
    test_file = './test_files/230309_001.frd'

    name = guess_experiment_name(test_file)
    assert name == '5th interaction with antibodie reg and imd - wt mut and wtDTT', "The guessed experiment name should be '5th interaction with antibodie reg and imd - wt mut and wtDTT'"

    test_file = './test_files/empty_file.frd'
    name = guess_experiment_name(test_file)
    assert name == 'Experiment'

def test_guess_experiment_type():

    non_existing_file = 'sarasaa'
    file_type = guess_experiment_type(non_existing_file)
    assert file_type == 'surface', "The default guessed experiment type should be 'surface'"

    mock_file = 'ExperimentStep.ini'
    file_type = guess_experiment_type(mock_file)
    assert file_type == 'surface', "The guessed experiment type should be 'surface'"

    mock_file = 'xxxx.frd'
    file_type = guess_experiment_type(mock_file)
    assert file_type == 'surface', "The guessed experiment type should be 'surface'"

    test_file = "./test_files/multi_cycle_kingenie.csv"
    file_type = guess_experiment_type(test_file)
    assert file_type == 'surface', "The guessed experiment type should be 'solution'"

    test_file = "./test_files/kingenie_solution.csv"
    file_type = guess_experiment_type(test_file)
    assert file_type == 'solution', "The guessed experiment type should be 'solution'"

def test_if_string_to_list():

    test_lst = if_string_to_list('test_string')
    assert isinstance(test_lst, list), "The output should be a list"

    test_lst = if_string_to_list(['test_string'])
    assert isinstance(test_lst, list), "The output should be a list"

    # check type error if input is not a string or a list
    with pytest.raises(TypeError):
        if_string_to_list(123)

def test_detect_time_list_continuos():

    assoc_time_lst = [ np.arange(0, 10, 1),  np.arange(20,30, 1)]
    disso_time_lst = [ np.arange(10, 20, 1)]

    result = detect_time_list_continuos(assoc_time_lst, disso_time_lst,tolerance=3)

    assert result == [True,True], "The first is True because it is the first association, the second is True because the association comes directly after the dissociation"

    assoc_time_lst = [ np.arange(0, 10, 1), np.arange(40, 50, 1)]
    disso_time_lst = [ np.arange(10, 20, 1)]

    result = detect_time_list_continuos(assoc_time_lst, disso_time_lst,tolerance=3)

    assert result == [True,False], "The first is True because it is the first association, the second is False because the association does not come directly after the dissociation"

def test_get_palette():
    # Test with a valid number of colors
    palette = get_palette(5)
    assert len(palette) == 5, "The palette should contain 5 colors"

    palette = get_palette(10)
    assert len(palette) == 10, "The palette should contain 10 colors"


    # Test with a number of colors greater than the available colors
    palette = get_palette(20)
    assert len(palette) == 20, "The palette should contain 20 colors even if it exceeds the default size"

def test_find_loading_column():

    # Create a mock DataFrame
    data = {
        'StepNumber': [1,1],
        'Type': ['hello', 'world']
    }

    df = pd.DataFrame(data)

    # Test with a valid step number
    loading_col = find_loading_column(df, step_number=1)
    assert loading_col is None, "There should be no loading column"

def test_subset_data():

    data = np.arange(200)

    subset = subset_data(data, max_points=100)

    assert len(subset) == 100, "The subset should contain 100 points"

def test_get_colors_from_numeric_values():

    values = [1,10,100]
    min_value = 0.1
    max_value = 1000

    colors = get_colors_from_numeric_values(values, min_value, max_value)

    assert len(colors) == len(values), "The number of colors should match the number of values" 
    
def test_sample_type_to_letter():

    assert sample_type_to_letter('SAMPLE') == 'S', "The word SAMPLE should map to the letter S"
    assert sample_type_to_letter('') == '', "An empty string should return an empty string"

def test_combine_sequences():

    seq1 = [1]
    seq2 = [3, 4]
    combined = combine_sequences(seq1, seq2)

    expected = [(1, 3), (1, 4)]
    np.testing.assert_array_equal(combined, expected, "The combined sequence should match the expected result")

def test_find_index_of_previous_step_type():

    # Create a mock DataFrame
    data = {
        'Type': ['BASELINE','ASSOC', 'DISASSOC','ASSOC', 'DISASSOC']
    }

    df_steps = pd.DataFrame(data)

    # Test with a valid step index and type
    step_index = 3
    step_type = 'DISASSOC'
    index = find_index_of_previous_step_type(df_steps, step_index, step_type)

    assert index == 1, "The index of the previous association should be 1"



import re
import math
import itertools

import numpy  as np
import pandas as pd

import matplotlib.colors as mcolors

from .palettes import set1_colors, set3_colors, VIRIDIS

__all__ = [
    'guess_experiment_name', 'etree_to_dict', 'combine_dicts','guess_experiment_type',
    'detect_time_list_continuos', 'find_loading_column', 'concat_signal_lst',
    'expand_parameter_list', 'combine_sequences', 'get_palette', 'get_plotting_df',
    'subset_data', 'sample_type_to_letter', 'get_colors_from_numeric_values',
    'if_string_to_list','find_index_of_previous_step_type','number_is_in_word']

def if_string_to_list(string_or_list):
    """
    Convert a string to a list if necessary.

    Parameters
    ----------
    string_or_list : str or list
        Input string or list to convert.

    Returns
    -------
    list
        Converted list.
        
    Raises
    ------
    TypeError
        If input is neither a string nor a list.
    """
    if isinstance(string_or_list, str):
        return [string_or_list]
    elif isinstance(string_or_list, list):
        return string_or_list
    else:
        raise TypeError("Input must be a string or a list of strings.")

def guess_experiment_name(frd_file):
    """
    Given a certain frd file, try to guess the experiment name.
    
    The frd file is the format exported by the Biacore software.

    Parameters
    ----------
    frd_file : str
        Path to the FRD file.
        
    Returns
    -------
    str
        Experiment name extracted from the file, or 'Experiment' if no name is found.
    """

    # Read the file content
    with open(frd_file, 'r') as file:
        content = file.read()

    # Define the regex pattern
    pattern = r'<ExperimentInfo Name="([^"]+)">'

    # Find the first match
    match = re.search(pattern, content)

    if match:
        experiment_name = match.group(1)  # Get the captured group

        # Remove any sequence of six numbers in a row (we assume it is a timestamp)
        experiment_name = re.sub(r'\b\d{6}\b', '', experiment_name)

        # Remove all extra spaces
        experiment_name = re.sub(r'\s+', ' ', experiment_name).strip()

        return experiment_name

    else:

        return 'Experiment'

def etree_to_dict(tree):
    """
    Converts XML tree to dictionary.
    
    Parameters
    ----------
    tree : ElementTree
        XML tree to convert.
        
    Returns
    -------
    dict
        Dictionary representation of the XML tree.
    """
    tree_dict = {}
    for elem in tree:
        tree_dict[elem.tag] = elem.text
    return tree_dict

def combine_dicts(list_dicts):
    """
    Combines dictionaries with the same keys into a dictionary with lists as values.
    
    Parameters
    ----------
    list_dicts : list
        List of dictionaries to combine.
        
    Returns
    -------
    dict
        Combined dictionary with lists as values.
        
    Notes
    -----
    All dictionaries must have the same keys.
    """
    new_dict = {}
    for key in list_dicts[0].keys():
        new_dict[key] = [one_dict[key] for one_dict in list_dicts]
    return new_dict

def guess_experiment_type(files):
    """
    Given file paths, try to guess if they correspond to surface-based or solution-based binding experiments.

    Parameters
    ----------
    files : str or list
        Path or list of paths to the experiment files.
        
    Returns
    -------
    str
        'surface' or 'solution' indicating the experiment type.
    """

    # convert to list if it is a string
    if isinstance(files, str):
        files = [files]

    for file in files:
        if file.endswith('.frd'):
            return 'surface'

    for file in files:
        if 'ExperimentStep.ini' in file:
            return 'surface'

    for file in files:
        if file.endswith('.csv'):

            # Find if we have a line with the protein concentration
            with open(file, 'r') as f:
                first_line = f.read().splitlines()[0]
                if 'Protein_concentration_micromolar' in first_line:
                    return 'solution'
                else:
                    return 'surface'

    return 'surface'

def detect_time_list_continuos(assoc_time_lst, disso_time_lst, tolerance=3):
    """
    Detect which association steps come directly after a dissociation step.
    
    Useful for single-cycle kinetics.

    Parameters
    ----------
    assoc_time_lst : list
        List of association time arrays.
    disso_time_lst : list
        List of dissociation time arrays.
    tolerance : float, optional
        Tolerance for the time difference (in seconds), default is 3.
        
    Returns
    -------
    list
        List of booleans indicating if the association phase has a dissociation phase just before.
    """

    continuos = []

    for i, element in enumerate(assoc_time_lst):

        if i == 0:

            continuos.append(element[0] < tolerance)

        else:

            prev_time = disso_time_lst[i-1][-1]
            continuos.append(element[0] < prev_time+tolerance)

    return continuos

def find_loading_column(df, step_number):
    """
    Find the loading column before the given step number.
    
    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing the columns 'Column', 'Type' and 'StepNumber'.
    step_number : int
        The step number to find the loading column before.
        
    Returns
    -------
    str or None
        The name of the loading column before the given step number, or None if no loading column is found.
    """

    # Filter the dataframe to get the loading columns before the given step number
    loading_columns = df[(df['Type'] == 'LOADING') & (df['StepNumber'] < step_number)]

    # If there are no loading columns, return None
    if loading_columns.empty:
        return None

    # Get the last loading column before the given step number
    last_loading_column = loading_columns.iloc[-1]

    # Return the column name
    return last_loading_column['Column']

def concat_signal_lst(signal_lst):
    """
    Concatenate a list of signals into a single array.

    Parameters
    ----------
    signal_lst : list
        List of signals to concatenate, each signal is a numpy array.

    Returns
    -------
    np.ndarray
        Concatenated signal.
    """

    return np.concatenate(signal_lst)

def expand_parameter_list(parameter_lst, id_list):
    """
    Expand a parameter list based on ID mapping.
    
    Given a list of n-parameters, such as [1,3] and another list
    containing IDs, such as [0,0,0,0,0,1,1],
    create a new list where the elements are repeated
    according to the IDs. In this case: [1,1,1,1,1,3,3]

    Parameters
    ----------
    parameter_lst : list
        List of n-Parameters.
    id_list : list
        List of m-IDs mapping to parameter indices.
        
    Returns
    -------
    list
        m-Parameters list expanded according to the IDs.
    """

    expanded_parameters = [parameter_lst[i] for i in id_list]
    return expanded_parameters

def combine_sequences(seq1, seq2):
    """
    Combine two sequences to generate all possible combinations of their elements.
    
    Parameters
    ----------
    seq1 : list
        First sequence of elements.
    seq2 : list
        Second sequence of elements.
        
    Returns
    -------
    list
        A list of tuples, where each tuple contains one element from seq1 and one from seq2.
    """
    return list(itertools.product(seq1, seq2))


def get_palette(n_colors):
    """
    Get a color palette with the specified number of colors.
    
    Parameters
    ----------
    n_colors : int
        Number of colors to include in the palette.
        
    Returns
    -------
    list
        List of hex color codes.
        
    Notes
    -----
    Uses Set1 from ColorBrewer for n_colors <= 9,
    Set3 for n_colors <= 12, and interpolated colors for n_colors > 12.
    """
    if n_colors <= 9:
        # Use Set1 from ColorBrewer
        return set1_colors[:n_colors]
    elif n_colors <= 12:
        # Use Set3 from ColorBrewer
        return set3_colors[:n_colors]
    else:

        # Create a colormap from the original palette
        cmap = mcolors.LinearSegmentedColormap.from_list("set3_interp", set3_colors)
        interpolated_colors = [mcolors.to_hex(cmap(i / (n_colors - 1))) for i in range(n_colors)]

        return interpolated_colors

def get_plotting_df(ID, labels=None):
    """
    Create a DataFrame for plotting purposes with colors and labels.
    
    Parameters
    ----------
    ID : list
        List of unique identifiers for each entry.
    labels : list, optional
        List of labels corresponding to each ID. Defaults to None, in which case the ID values are used as labels.
        
    Returns
    -------
    pd.DataFrame
        DataFrame containing Internal_ID, Color, Legend, and Show columns.
    """
    if labels is None:
        labels = ID
    n_total = len(labels)
    df = pd.DataFrame({
        "Internal_ID": ID,
        "Color": get_palette(n_total),
        "Legend": labels,
        "Show": [True] * n_total
    })
    return df

def subset_data(data, max_points=200):
    """
    Subset data to a maximum number of points for plotting.
    
    Parameters
    ----------
    data : list or np.ndarray
        Data to be subsetted.
    max_points : int, optional
        Maximum number of points to retain, default is 200.
        
    Returns
    -------
    list or np.ndarray
        Subsetted data with at most max_points points.
    """
    total_points = len(data)
    if total_points > max_points:
        step = math.ceil(total_points / max_points)
        data = data[::step]
    return data

def sample_type_to_letter(string):
    """
    Convert sample type string to a single letter code.
    
    Parameters
    ----------
    string : str
        Sample type string.
        
    Returns
    -------
    str
        Single letter code representing the sample type.
        
    Notes
    -----
    Mapping:
    - KSAMPLE/SAMPLE -> S
    - LOADING/KLOADING -> L
    - NEUTRALIZATION -> N
    - REGENERATION -> R
    - BASELINE/BUFFER -> B
    - Others -> First letter of the string
    """

    mapping = {
        "KSAMPLE": "S",
        "SAMPLE": "S",
        "LOADING": "L",
        "KLOADING": "L",
        "NEUTRALIZATION": "N",
        "REGENERATION": "R",
        "BASELINE": "B",
        "BUFFER": "B"
    }
    if string not in mapping:
        return string[0] if string else ""
    return mapping[string]

def get_colors_from_numeric_values(values, min_val, max_val, use_log_scale=True):
    """
    Map numeric values to colors in the VIRIDIS palette based on a specified range.
    
    Parameters
    ----------
    values : list or np.ndarray
        Numeric values to map to colors.
    min_val : float
        Minimum value of the range.
    max_val : float
        Maximum value of the range.
    use_log_scale : bool, optional
        Whether to use logarithmic scaling for the values, default is True.
        
    Returns
    -------
    list
        List of hex color codes corresponding to the input values.
    """
    values = np.array(values)
    if use_log_scale:
        min_val = np.log10(min_val)
        max_val = np.log10(max_val)
        values = np.log10(values)
    seq = np.linspace(min_val, max_val, 21)
    idx = [np.argmin(np.abs(v - seq)) for v in values]
    return [VIRIDIS[i] for i in idx]

def find_index_of_previous_step_type(df_steps, step_index, step_type):
    """
    Find the index of the first step of a given type before a specified step index.
    
    Parameters
    ----------
    df_steps : pd.DataFrame
        DataFrame containing step information with a 'Type' column.
    step_index : int
        Current step index.
    step_type : str
        Type of step to search for.
        
    Returns
    -------
    int or None
        Index of the first step of the specified type before the given step index,
        or None if no matching step is found.
    """
    i = None
    # Find the index of the first step_type, from the single cycle
    # Iterate over the previous steps, two at a time, until we find a step that is not a step of the same type

    for idx in range(step_index-2, 0, -2):
        if df_steps['Type'][idx] != step_type:
            i = idx
            break

    return i


import numpy as np
import pytest
import os

from pykingenie.utils.math import single_exponential, double_exponential, median_filter, rss_p, get_rss, get_desired_rss

def test_single_exponential():
    t = np.array([0, 1, 2, 3, 4])
    a0 = 1.0
    a1 = 2.0
    kobs = 0.5

    result = single_exponential(t, a0, a1, kobs)
    expected = a0 + a1 * np.exp(-kobs * t)

    np.testing.assert_array_almost_equal(result, expected)

def test_double_exponential():
    t = np.array([0, 1, 2, 3, 4])
    a0 = 1.0
    a1 = 2.0
    kobs1 = 0.5
    a2 = 3.0
    kobs2 = 0.3

    result = double_exponential(t, a0, a1, kobs1, a2, kobs2)
    expected = a0 + a1 * np.exp(-kobs1 * t) + a2 * np.exp(-kobs2 * t)

    np.testing.assert_array_almost_equal(result, expected)

def test_median_filter():
    y = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    x = np.arange(len(y)) / 100

    rolling_window = 3 / 100

    filtered_y = median_filter(y, x, rolling_window)

    # Check if the filtered_y is a numpy array
    assert isinstance(filtered_y, np.ndarray), "The filtered_y should be a numpy array."

    # Check the length of the filtered_y
    assert len(filtered_y) == len(y), "The length of filtered_y should match the length of y."

    expected_values_after_padding = np.array([2, 3, 4, 5, 6, 7, 8])

    # Check if the filtered values match the expected values
    np.testing.assert_array_almost_equal(filtered_y[2:], expected_values_after_padding)

def test_rss_p():

    rrs0 = 100
    n = 10
    p = 5
    alfa = 0.05

    result = rss_p(rrs0, n, p, alfa)

    np.testing.assert_almost_equal(result, 232.1578194)

def test_get_rss():

    y     = np.arange(10)
    y_fit = y + 1

    rss_expected = 10

    rss = get_rss(y, y_fit)

    np.testing.assert_equal(rss, rss_expected)

def test_get_desired_rss():

    y = np.arange(10)
    y_fit = y + 1
    n = len(y)
    p = 2
    alpha=0.05

    rss = get_desired_rss(y, y_fit, n, p, alpha)

    np.testing.assert_almost_equal(rss, 16.647068,decimal=5)



import numpy as np
import pytest

from numpy.testing import assert_allclose

from pykingenie.utils.fitting_general import (
    convert_to_numpy_array,
    fit_single_exponential
)

def test_convert_to_numpy_array():

    l = [1, 2, 3]
    arr = convert_to_numpy_array(l)
    assert isinstance(arr, np.ndarray), "The output should be a numpy array"

    l = np.array([1, 2, 3])
    arr = convert_to_numpy_array(l)
    assert isinstance(arr, np.ndarray), "The output should be a numpy array"

def test_fit_single_exponential():

    t = np.arange(0, 5, 0.1)
    k_obs = 5
    a0 = 0.5
    a1 = 10
    y = a0 + a1 * np.exp(-k_obs * t)

    fit_params, cov, fit_y = fit_single_exponential(y,t)

    assert_allclose(fit_params, [a0,a1,k_obs], rtol=0.01)